"""DICOM dataset utility functions."""

from io import BytesIO
import logging
from pathlib import Path
from typing import Optional, List, Tuple
import zlib

from pydicom import Dataset
from pydicom.dataset import FileMetaDataset
from pydicom.dataelem import DataElement
from pydicom.filebase import DicomBytesIO
from pydicom.filereader import read_dataset, read_preamble
from pydicom.filewriter import write_dataset
from pydicom.tag import BaseTag
from pydicom.uid import UID

from pynetdicom import PYNETDICOM_IMPLEMENTATION_UID, PYNETDICOM_IMPLEMENTATION_VERSION
from pynetdicom.utils import pretty_bytes


LOGGER = logging.getLogger("pynetdicom.dsutils")


def create_file_meta(
    *,
    sop_class_uid: UID,
    sop_instance_uid: UID,
    transfer_syntax: UID,
    implementation_uid: UID = PYNETDICOM_IMPLEMENTATION_UID,
    implementation_version: str = PYNETDICOM_IMPLEMENTATION_VERSION,
) -> FileMetaDataset:
    """Return a new file meta dataset

    .. versionadded:: 2.0

    Parameters
    ----------
    sop_class_uid : pydicom.uid.UID
        The value for the *Media Storage SOP Class UID*.
    sop_instance_uid : pydicom.uid.UID
        The value for the *Media Storage SOP Instance UID*.
    transfer_syntax : pydicom.uid.UID
        The value for the *Transfer Syntax UID*.
    implementation_uid : pydicom.uid.UID, optional
        The value for the *Implementation Class UID*.
    implementation_version : str, optional
        The value for the *Implementation Version Name*.

    Returns
    -------
    pydicom.dataset.FileMetaDataset
        The File Meta dataset
    """
    file_meta = FileMetaDataset()

    file_meta.FileMetaInformationGroupLength = 0
    file_meta.FileMetaInformationVersion = b"\x00\x01"
    file_meta.MediaStorageSOPClassUID = sop_class_uid
    file_meta.MediaStorageSOPInstanceUID = sop_instance_uid
    file_meta.TransferSyntaxUID = transfer_syntax
    file_meta.ImplementationClassUID = implementation_uid
    file_meta.ImplementationVersionName = implementation_version

    # File Meta Information is always encoded as Explicit VR Little Endian
    file_meta.is_little_endian = True
    file_meta.is_implicit_VR = False

    return file_meta


def decode(
    bytestring: BytesIO,
    is_implicit_vr: bool,
    is_little_endian: bool,
    deflated: bool = False,
) -> Dataset:
    """Decode `bytestring` to a *pydicom* :class:`~pydicom.dataset.Dataset`.

    .. versionchanged:: 1.5

        Added `deflated` keyword parameter

    Parameters
    ----------
    byestring : io.BytesIO
        The encoded dataset in the DIMSE Message sent from the peer AE.
    is_implicit_vr : bool
        The dataset is encoded as implicit (``True``) or explicit VR
        (``False``).
    is_little_endian : bool
        The byte ordering of the encoded dataset, ``True`` for little endian,
        ``False`` for big endian.
    deflated : bool, optional
        ``True`` if the dataset has been encoded using *Deflated Explicit VR
        Little Endian* transfer syntax (default ``False``).

    Returns
    -------
    pydicom.dataset.Dataset
        The decoded dataset.
    """
    transfer_syntax = ""
    if deflated:
        transfer_syntax = "Deflated "
        is_implicit_vr = False
        is_little_endian = True

    transfer_syntax += "Little Endian" if is_little_endian else "Big Endian"
    if is_implicit_vr:
        transfer_syntax += " Implicit"
    else:
        transfer_syntax += " Explicit"

    LOGGER.debug('pydicom.read_dataset() TransferSyntax="%s"', transfer_syntax)

    # Rewind to the start of the stream
    bytestring.seek(0)

    if deflated:
        # Decompress the dataset
        bytestring = BytesIO(zlib.decompress(bytestring.getvalue(), -zlib.MAX_WBITS))
        bytestring.seek(0)

    # Decode the dataset
    return read_dataset(bytestring, is_implicit_vr, is_little_endian)


def encode(
    ds: Dataset, is_implicit_vr: bool, is_little_endian: bool, deflated: bool = False
) -> Optional[bytes]:
    """Encode a *pydicom* :class:`~pydicom.dataset.Dataset` `ds`.

    .. versionchanged:: 1.5

        Added `deflated` keyword parameter

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset to encode
    is_implicit_vr : bool
        The element encoding scheme the dataset will be encoded with, ``True``
        for implicit VR, ``False`` for explicit VR.
    is_little_endian : bool
        The byte ordering the dataset will be encoded in, ``True`` for little
        endian, ``False`` for big endian.
    deflated : bool, optional
        ``True`` if the dataset is to be encoded using *Deflated Explicit VR
        Little Endian* transfer syntax (default ``False``).

    Returns
    -------
    bytes or None
        The encoded dataset as :class:`bytes` (if successful) or ``None`` if
        the encoding failed.
    """
    # pylint: disable=broad-except
    fp = DicomBytesIO()
    fp.is_implicit_VR = is_implicit_vr
    fp.is_little_endian = is_little_endian
    try:
        write_dataset(fp, ds)
    except Exception as exc:
        LOGGER.error("pydicom.write_dataset() failed:")
        LOGGER.exception(exc)
        fp.close()
        return None

    bytestring: bytes = fp.parent.getvalue()  # type: ignore
    fp.close()

    if deflated:
        # Compress the encoded dataset
        compressor = zlib.compressobj(
            zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, -zlib.MAX_WBITS
        )
        bytestring = compressor.compress(bytestring)
        bytestring += compressor.flush()
        bytestring += b"\x00" if len(bytestring) % 2 else b""

    return bytestring


def pretty_dataset(ds: Dataset, indent: int = 0, indent_char: str = "  ") -> List[str]:
    """Return a list of pretty dataset strings.

    .. versionadded:: 1.5

    Parameters
    ----------
    ds : pydicom.dataset.Dataset
        The dataset to beautify.
    indent : int, optional
        The indentation level of the current dataset (default: ``0``).
    indent_char : str, optional
        The character(s) to use when indenting the dataset (default ``'  '``).

    Returns
    -------
    list of str
    """
    out = []
    for elem in iter(ds):
        if elem.VR == "SQ":
            out.append(pretty_element(elem))
            for ii, item in enumerate(elem.value):
                msg = f"(Sequence item #{ii + 1})"
                out.append(indent_char * (indent + 1) + msg)
                out.extend(pretty_dataset(item, indent + 2))
        else:
            out.append(indent_char * indent + pretty_element(elem))

    return out


def pretty_element(elem: DataElement) -> str:
    """Return a pretty element string.

    .. versionadded:: 1.5

    Parameters
    ----------
    elem : pydicom.dataelem.DataElement
        The element to beautify.

    Returns
    -------
    str
    """
    try:
        value = elem.value
        if elem.VM == 0 and elem.VR != "SQ":
            # Empty value
            value = "(no value available)"
        elif elem.VR in ["OB", "OD", "OF", "OL", "OW", "OV"]:
            # Byte VRs
            if elem.VM == 1:
                # Single value
                length = len(elem.value)
                if length <= 13:
                    value = pretty_bytes(elem.value, prefix="", delimiter=" ")
                    value = f"[{value[0]}]"
                else:
                    value = f"({len(elem.value)} bytes of binary data)"
            else:
                # Multiple values - probably non-conformant
                total_length = sum([len(ii) for ii in elem.value])
                value = f"({total_length} bytes of binary data)"
        elif elem.VR != "SQ":
            # Non-sequence elements
            if elem.VM == 1:
                value = f"[{elem.value}]"
            else:
                value = "\\".join([str(ii) for ii in elem.value])
                value = f"[{value}]"
        elif elem.VR == "SQ":
            # Sequence elements
            if elem.VM == 1:
                value = f"(Sequence with {len(elem.value)} item)"
            else:
                value = f"(Sequence with {len(elem.value)} items)"

    except Exception as exc:
        value = "(pynetdicom failed to beautify value)"

    return "({:04X},{:04X}) {} {: <40} # {} {}".format(
        elem.tag.group, elem.tag.element, elem.VR, value, elem.VM, elem.keyword
    )


def split_dataset(path: Path) -> Tuple[Dataset, int]:
    """Return the file meta elements and the offset to the start of the dataset

    .. versionadded:: 2.0

    Parameters
    ----------
    path : pathlib.Path
        The path to a dataset written in the DICOM File Format.

    Returns
    -------
    pydicom.dataset.Dataset, int
        The File Meta elements as a Dataset instance and the byte offset to
        the start of the dataset itself. The File Meta dataset may be empty if
        no File Meta is present.
    """

    def _not_group_0002(tag: BaseTag, VR: Optional[str], length: int) -> bool:
        """Return True if the tag is not in group 0x0002, False otherwise."""
        return tag.group != 2

    with open(path, "rb") as fp:
        read_preamble(fp, False)
        file_meta = read_dataset(
            fp, is_implicit_VR=False, is_little_endian=True, stop_when=_not_group_0002
        )
        return file_meta, fp.tell()



"""Implementation of the Presentation service."""

import logging
from typing import (
    Union,
    Optional,
    List,
    Any,
    TYPE_CHECKING,
    NamedTuple,
    cast,
    Dict,
    Tuple,
)

from pydicom.uid import UID

from pynetdicom._globals import DEFAULT_TRANSFER_SYNTAXES
from pynetdicom import sop_class as SOP_CLASS_MODULE
from pynetdicom.sop_class import (
    _APPLICATION_EVENT_CLASSES,
    _BASIC_WORKLIST_CLASSES,
    _COLOR_PALETTE_CLASSES,
    _DEFINED_PROCEDURE_CLASSES,
    _DISPLAY_SYSTEM_CLASSES,
    _HANGING_PROTOCOL_CLASSES,
    _IMPLANT_TEMPLATE_CLASSES,
    _INSTANCE_AVAILABILITY_CLASSES,
    _MEDIA_CREATION_CLASSES,
    _MEDIA_STORAGE_CLASSES,
    _NON_PATIENT_OBJECT_CLASSES,
    _PRINT_MANAGEMENT_CLASSES,
    _PROCEDURE_STEP_CLASSES,
    _PROTOCOL_APPROVAL_CLASSES,
    _QR_CLASSES,
    _RELEVANT_PATIENT_QUERY_CLASSES,
    _RT_MACHINE_VERIFICATION_CLASSES,
    _STORAGE_CLASSES,
    _STORAGE_COMMITMENT_CLASSES,
    _SUBSTANCE_ADMINISTRATION_CLASSES,
    _UNIFIED_PROCEDURE_STEP_CLASSES,
    _VERIFICATION_CLASSES,
)
from pynetdicom.utils import validate_uid, set_uid

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation


LOGGER = logging.getLogger("pynetdicom.presentation")


# Used with the event handlers to give the users access to the context
class PresentationContextTuple(NamedTuple):
    """:func:`namedtuple<collections.namedtuple>` representation of an accepted
    :class:`PresentationContext`.
    """

    context_id: int
    abstract_syntax: UID
    transfer_syntax: UID


DEFAULT_ROLE = (True, False, False, True)
BOTH_SCU_SCP_ROLE = (True, True, True, True)
CONTEXT_REJECTED = (False, False, False, False)
INVERTED_ROLE = (False, True, True, False)
_RoleType = Dict[
    Tuple[Optional[bool], Optional[bool]],
    Dict[Tuple[Optional[bool], Optional[bool]], Tuple[bool, bool, bool, bool]],
]
SCP_SCU_ROLES: _RoleType = {
    # (Requestor role, Acceptor role) : Outcome
    # No SCP/SCU Role Selection proposed
    (None, None): {
        (None, None): DEFAULT_ROLE,
        (None, True): DEFAULT_ROLE,
        (None, False): DEFAULT_ROLE,
        (True, None): DEFAULT_ROLE,
        (False, None): DEFAULT_ROLE,
        (True, True): DEFAULT_ROLE,
        (True, False): DEFAULT_ROLE,
        (False, False): DEFAULT_ROLE,
        (False, True): DEFAULT_ROLE,
    },
    (True, True): {
        (None, None): DEFAULT_ROLE,
        (None, True): DEFAULT_ROLE,
        (None, False): DEFAULT_ROLE,
        (True, None): DEFAULT_ROLE,
        (False, None): DEFAULT_ROLE,
        (True, True): BOTH_SCU_SCP_ROLE,
        (True, False): DEFAULT_ROLE,
        (False, False): CONTEXT_REJECTED,
        (False, True): INVERTED_ROLE,
    },
    (True, False): {
        (None, None): DEFAULT_ROLE,
        (None, True): DEFAULT_ROLE,
        (None, False): DEFAULT_ROLE,
        (True, None): DEFAULT_ROLE,
        (False, None): DEFAULT_ROLE,
        (True, True): DEFAULT_ROLE,  # Invalid
        (True, False): DEFAULT_ROLE,
        (False, False): CONTEXT_REJECTED,
        (False, True): CONTEXT_REJECTED,  # Invalid
    },
    (False, True): {
        (None, None): DEFAULT_ROLE,
        (None, True): DEFAULT_ROLE,
        (None, False): DEFAULT_ROLE,
        (True, None): DEFAULT_ROLE,
        (False, None): DEFAULT_ROLE,
        (True, True): INVERTED_ROLE,  # Invalid
        (True, False): CONTEXT_REJECTED,  # Invalid
        (False, False): CONTEXT_REJECTED,
        (False, True): INVERTED_ROLE,
    },
    # False, False proposed x
    (False, False): {
        (None, None): DEFAULT_ROLE,
        (None, True): DEFAULT_ROLE,
        (None, False): DEFAULT_ROLE,
        (True, None): DEFAULT_ROLE,
        (False, None): DEFAULT_ROLE,
        (True, True): CONTEXT_REJECTED,  # Invalid
        (True, False): CONTEXT_REJECTED,  # Invalid
        (False, False): CONTEXT_REJECTED,
        (False, True): CONTEXT_REJECTED,  # Invalid
    },
}


class PresentationContext:
    """A Presentation Context primitive.

    **Rules**

    - Each Presentation Context (request) contains:

      - One context ID, an odd integer between 1 and 255.
      - One abstract syntax.
      - One or more transfer syntaxes.
    - Each Presentation Context (response) contains:

      - One context ID, corresponding to a Presentation Context received from
        the *Requestor*
      - A result, one of ``0x00`` (acceptance), ``0x01`` (user rejection),
        ``0x02`` (provider rejection), ``0x03`` (abstract syntax not supported)
        or ``0x04`` (transfer syntaxes not supported).
      - If the result is ``0x00``, then a transfer syntax.
      - If any other result, then a transfer syntax may or may not be present.
    - If the result is not ``0x00`` then the transfer syntax in the reply is
      not significant.
    - The same abstract syntax can be present in more than one Presententation
      Context.
    - Only one transfer syntax can be accepted per Presentation Context.
    - The Presentation Contexts may be sent by the *Requestor* in any order.
    - The Presentation Contexts may be sent by the *Acceptor* in any order.

    **SCP/SCU Role Selection Negotiation**

    - If no role selection negotiation then the *Requestor* is SCU and the
      *Acceptor* is SCP
    - The association *Acceptor* cannot accept a role that has not been
      proposed (i.e. cannot return 1 when the proposed value is 0).
    - The association *Requestor* may be SCP only, SCU only or both SCU and
      SCP.

    +---------------------+---------------------+-------------------+---------+
    | Requestor           | Acceptor            | Outcome           | Notes   |
    +----------+----------+----------+----------+---------+---------+         |
    | scu_role | scp_role | scu_role | scp_role | Req.    | Acc.    |         |
    +==========+==========+==========+==========+=========+=========+=========+
    | N/A      | N/A      | N/A      | N/A      | SCU     | SCP     | Default |
    +----------+----------+----------+----------+---------+---------+---------+
    | True     | True     | False    | False    | N/A     | N/A     | Rejected|
    |          |          |          +----------+---------+---------+---------+
    |          |          |          | True     | SCP     | SCU     |         |
    |          |          +----------+----------+---------+---------+---------+
    |          |          | True     | False    | SCU     | SCP     | Default |
    |          |          |          +----------+---------+---------+---------+
    |          |          |          | True     | SCU/SCP | SCU/SCP |         |
    +----------+----------+----------+----------+---------+---------+---------+
    | True     | False    | False    | False    | N/A     | N/A     | Rejected|
    |          |          +----------+          +---------+---------+---------+
    |          |          | True     |          | SCU     | SCP     | Default |
    +----------+----------+----------+----------+---------+---------+---------+
    | False    | True     | False    | False    | N/A     | N/A     | Rejected|
    |          |          |          +----------+---------+---------+---------+
    |          |          |          | True     | SCP     | SCU     |         |
    +----------+----------+----------+----------+---------+---------+---------+
    | False    | False    | False    | False    | N/A     | N/A     | Rejected|
    +----------+----------+----------+----------+---------+---------+---------+

    As can be seen from the above table there are four possible outcomes:

    * *Requestor* is SCU, *Acceptor* is SCP (default roles)
    * *Requestor* is SCP, *Acceptor* is SCU
    * *Requestor* and *Acceptor* are both SCU/SCP
    * *Requestor* and *Acceptor* are neither (context rejected)

    Attributes
    ---------
    result : int or None
        If part of an A-ASSOCIATE (request) then ``None``. If part of an
        A-ASSOCIATE (response) then one of ``0x00``, ``0x01``, ``0x02``,
        ``0x03``, ``0x04``.

    References
    ----------

    * DICOM Standard, Part 7, Annexes :dcm:`D.3.2<part07.html#sect_D.3.2>`
      and :dcm:`D.3.3.4<part07.html#sect_D.3.3.4>`
    * DICOM Standard, Part 8, Sections :dcm:`9.3.2.2
      <part08.html#sect_9.3.2.2>`, :dcm:`9.3.3.2 <part08.html#sect_9.3.3.2>`
      and :dcm:`Annex B <part08.html#chapter_B>`
    """

    def __init__(self) -> None:
        """Create a new object."""
        self._context_id: Optional[int] = None
        self._abstract_syntax: Optional[UID] = None
        self._transfer_syntax: List[UID] = []
        self.result: Optional[int] = None

        # Used with SCP/SCU Role Selection negotiation
        self._scu_role: Optional[bool] = None
        self._scp_role: Optional[bool] = None

        # Used to track the allowed use of the context
        self._as_scp: Optional[bool] = None
        self._as_scu: Optional[bool] = None

    @property
    def abstract_syntax(self) -> Optional[UID]:
        """Get or set the context's *Abstract Syntax* as
        :class:`~pydicom.uid.UID`.

        Parameters
        ----------
        value : str or bytes or pydicom.uid.UID
            The abstract syntax UID.
        """
        return self._abstract_syntax

    @abstract_syntax.setter
    def abstract_syntax(self, value: Union[str, bytes, UID]) -> None:
        """Set the context's *Abstract Syntax*."""
        self._abstract_syntax = set_uid(value, "abstract_syntax", True, False, True)

    def add_transfer_syntax(self, syntax: Union[None, str, bytes, UID]) -> None:
        """Append a transfer syntax to the presentation context.

        Parameters
        ----------
        syntax : pydicom.uid.UID, bytes or str
            The transfer syntax to add to the presentation context.
        """
        if syntax is None:
            return

        if isinstance(syntax, str):  # includes UID
            syntax = UID(syntax)
        elif isinstance(syntax, bytes):
            syntax = UID(syntax.decode("ascii"))
        else:
            LOGGER.error("Attempted to add an invalid transfer syntax")
            return

        if syntax is not None and not validate_uid(syntax):
            LOGGER.error("'transfer_syntax' contains an invalid UID")
            raise ValueError("'transfer_syntax' contains an invalid UID")

        if syntax and not syntax.is_valid:
            LOGGER.warning(f"The Transfer Syntax Name '{syntax}' is non-conformant")

        # If the transfer syntax is rejected we may add an empty str
        if syntax not in self._transfer_syntax and syntax != "":
            if not syntax.is_valid:
                LOGGER.warning(
                    "A non-conformant UID has been added to 'transfer_syntax'"
                )

            if not syntax.is_private and not syntax.is_transfer_syntax:
                LOGGER.warning(
                    "A UID has been added to 'transfer_syntax' that is not a "
                    "transfer syntax"
                )

            self._transfer_syntax.append(syntax)

    @property
    def as_scp(self) -> Optional[bool]:
        """Return ``True`` if can act as an SCP for the context.

        If ``True`` then the association *Acceptor* can act as SCP for the
        current context, otherwise it cannot. A non-``None`` value is only
        available after association negotiation has been completed.
        """
        return self._as_scp

    @property
    def as_scu(self) -> Optional[bool]:
        """Return ``True`` if can act as an SCU for the context.

        If ``True`` then the association *Acceptor* can act as SCU for the
        current context, otherwise it cannot. A non-``None`` value is only
        available after association negotiation has been completed.
        """
        return self._as_scu

    @property
    def as_tuple(self) -> PresentationContextTuple:
        """Return a :func:`namedtuple<collections.namedtuple>` representation
        of the presentation context.

        Intended to be used when the result is ``0x00`` (accepted) as only the
        first transfer syntax item is returned in the tuple.

        Returns
        -------
        PresentationContextTuple
            A representation of an accepted presentation context.
        """
        return PresentationContextTuple(
            cast(int, self.context_id),
            cast(UID, self.abstract_syntax),
            self.transfer_syntax[0],
        )

    @property
    def context_id(self) -> Optional[int]:
        """Return the context's *ID* parameter as an :class:`int`."""
        return self._context_id

    @context_id.setter
    def context_id(self, value: Optional[int]) -> None:
        """Set the context's *ID* parameter.

        Parameters
        ----------
        value : int
            An odd integer between 1 and 255 (inclusive).
        """
        if value is not None and (not 1 <= value <= 255 or value % 2 == 0):
            raise ValueError(
                "'context_id' must be an odd integer between 1 and 255, inclusive"
            )

        self._context_id = value

    def __eq__(self, other: Any) -> bool:
        """Return ``True`` if `self` is equal to `other`."""
        if self is other:
            return True

        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__

        return NotImplemented

    def __hash__(self) -> int:
        """Return a hash of the context."""
        return hash(
            (
                self.abstract_syntax,
                self.context_id,
                tuple(self.transfer_syntax),
                self.as_scp,
                self.as_scu,
            )
        )

    def __ne__(self, other: Any) -> bool:
        """Return ``True`` if `self` does not equal `other`."""
        return not self == other

    def __repr__(self) -> str:
        """Representation of the Presentation Context."""
        return cast(UID, self.abstract_syntax).name

    @property
    def scp_role(self) -> Optional[bool]:
        """Get or set if a proposed SCP role will be accepted.

        Parameters
        ----------
        val : bool or None
            If ``True`` then accept if the association *Requestor* proposes
            the SCP role for itself, ``False`` to reject the proposal. If
            ``None`` (default) then no SCP/SCU Role Selection reply will be
            sent. If either of :attr:`scu_role` or :attr:`scp_role` is ``None``
            then both will assumed to be.
        """
        return self._scp_role

    @scp_role.setter
    def scp_role(self, val: Optional[bool]) -> None:
        """Set whether to accept the proposed SCP role (as *Acceptor*)."""
        if not isinstance(val, (bool, type(None))):
            raise TypeError("'scp_role' must be a bool or None")

        self._scp_role = val

    @property
    def scu_role(self) -> Optional[bool]:
        """Get or set if a proposed SCU role will be accepted.

        Parameters
        ----------
        val : bool or None
            If ``True`` then accept if the association *Requestor* proposes
            the SCU role for itself, ``False`` to reject the proposal. If
            ``None`` (default) then no SCP/SCU Role Selection reply will be
            sent. If either of :attr:`scu_role` or :attr:`scp_role` is ``None``
            then both will assumed to be.
        """
        return self._scu_role

    @scu_role.setter
    def scu_role(self, val: Optional[bool]) -> None:
        """Set whether to accept the proposed SCU role (as *Acceptor*)."""
        if not isinstance(val, (bool, type(None))):
            raise TypeError("'scu_role' must be a bool or None")

        self._scu_role = val

    @property
    def status(self) -> str:
        """Return a descriptive :class:`str` of the context's *Result*.

        Returns
        -------
        str
            The string representation of the negotiated result.
        """
        s = {
            None: "Pending",
            0x00: "Accepted",
            0x01: "User Rejected",
            0x02: "Provider Rejected",
            0x03: "Abstract Syntax Not Supported",
            0x04: "Transfer Syntax(es) Not Supported",
        }
        try:
            return s[self.result]
        except KeyError:
            return "Unknown"

    def __str__(self) -> str:
        """String representation of the Presentation Context."""
        s = []
        if self.context_id is not None:
            s.append(f"ID: {self.context_id}")

        if self.abstract_syntax is not None:
            s.append(f"Abstract Syntax: {self.abstract_syntax.name}")

        s.append("Transfer Syntax(es):")
        if not self.transfer_syntax:
            s.append("    (none)")
        else:
            s.extend(f"    ={ts.name}" for ts in self.transfer_syntax)

        if self.result is not None:
            s.append(f"Result: {self.status}")

        if None not in (self.as_scu, self.as_scp):
            if self.as_scu and not self.as_scp:
                s.append("Role: SCU only")
            elif self.as_scu and self.as_scp:
                s.append("Role: SCU and SCP")
            elif not self.as_scu and self.as_scp:
                s.append("Role: SCP only")
            else:
                s.append("Role: (none)")

        return "\n".join(s)

    @property
    def transfer_syntax(self) -> List[UID]:
        """Get or set the context's *Transfer Syntaxes* as a :class:`list`.

        Returns
        -------
        list of pydicom.uid.UID
            The context's transfer syntaxes.
        """
        return self._transfer_syntax

    @transfer_syntax.setter
    def transfer_syntax(self, syntaxes: List[Union[str, bytes, UID]]) -> None:
        """Set the context's *Transfer Syntaxes*.

        Parameters
        ----------
        syntaxes : list of (str or bytes or pydicom.uid.UID)
            The transfer syntax UIDs to add to the Presentation Context.
        """
        if not isinstance(syntaxes, list):
            raise TypeError("'transfer_syntax' must be a list")

        self._transfer_syntax = []

        for syntax in syntaxes:
            self.add_transfer_syntax(syntax)


RoleType = Optional[Dict[UID, Tuple[Optional[bool], Optional[bool]]]]
ListCXType = List[PresentationContext]
CXNegotiationReturn = Tuple[
    List[PresentationContext], List["SCP_SCU_RoleSelectionNegotiation"]
]


def negotiate_unrestricted(
    rq_contexts: ListCXType, ac_contexts: ListCXType, roles: RoleType = None
) -> CXNegotiationReturn:
    """Process the Presentation Contexts as an Association *Acceptor*
    with an unrestricted storage service.

    ..versionadded:: 2.0

    Parameters
    ----------
    rq_contexts : list of PresentationContext
        The Presentation Contexts proposed by the peer. Each item has
        values for Context ID, Abstract Syntax and Transfer Syntax.
    ac_contexts : list of PresentationContext
        The Presentation Contexts supported by the local AE when acting
        as an Association *Acceptor*. Each item has values for Context ID,
        Abstract Syntax and Transfer Syntax.
    roles : dict or None
        If the *Requestor* has included one or more SCP/SCU Role Selection
        Negotiation items then this will be a :class:`dict` of
        ``{'SOP Class UID' : (SCU role, SCP role)}``, otherwise ``None``
        (default)

    Returns
    -------
    list of PresentationContext
        The accepted presentation context items, each with a Result value
        a Context ID, an Abstract Syntax and one Transfer Syntax item.
        Items are sorted in increasing Context ID value.
    list of SCP_SCU_RoleSelectionNegotiation
        If `roles` is not ``None`` then this is a :class:`list` of SCP/SCU Role
        Selection Negotiation items that can be sent back to the *Requestor*.
    """
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

    roles = roles or {}
    storage_contexts: List[PresentationContext] = []
    non_storage_contexts: List[PresentationContext] = []
    reply_roles: Dict[UID, SCP_SCU_RoleSelectionNegotiation] = {}
    storage_uids = _STORAGE_CLASSES.values()

    # Split out private/unknown/storage cx's from everything else
    for cx in rq_contexts:
        ab_syntax = cast(UID, cx.abstract_syntax)
        if (
            ab_syntax.is_private
            or ab_syntax in storage_uids
            or not hasattr(SOP_CLASS_MODULE, ab_syntax.keyword)
        ):
            storage_contexts.append(cx)
        else:
            non_storage_contexts.append(cx)

    # Negotiate non-storage contexts as normal
    result_cx, result_roles = negotiate_as_acceptor(
        non_storage_contexts, ac_contexts, roles
    )

    # Accept all storage-like contexts
    for rcx in storage_contexts:
        cx = PresentationContext()
        cx.context_id = rcx.context_id
        cx.abstract_syntax = rcx.abstract_syntax
        cx.transfer_syntax = [rcx.transfer_syntax[0]]
        cx.result = 0x00
        cx._as_scu = True
        cx._as_scp = True

        # Role selection
        if rcx.abstract_syntax in roles:
            role = SCP_SCU_RoleSelectionNegotiation()
            role.sop_class_uid = rcx.abstract_syntax

            rq_roles = roles[rcx.abstract_syntax]
            outcome = SCP_SCU_ROLES[rq_roles][(True, True)]
            cx._as_scu = outcome[2]
            cx._as_scp = outcome[3]

            # Can't return 0x01 if proposed 0x00
            role.scu_role = False if not rq_roles[0] else True
            role.scp_role = False if not rq_roles[1] else True

            reply_roles[cast(UID, cx.abstract_syntax)] = role

        result_cx.append(cx)

    # Not required but a nice thing to do
    result_cx = sorted(result_cx, key=lambda x: cast(int, x.context_id))
    result_roles = sorted(
        reply_roles.values(), key=lambda x: cast(UID, x.sop_class_uid)
    )

    return result_cx, result_roles


def negotiate_as_acceptor(
    rq_contexts: ListCXType, ac_contexts: ListCXType, roles: RoleType = None
) -> CXNegotiationReturn:
    """Process the Presentation Contexts as an Association *Acceptor*.

    Parameters
    ----------
    rq_contexts : list of PresentationContext
        The Presentation Contexts proposed by the peer. Each item has
        values for Context ID, Abstract Syntax and Transfer Syntax.
    ac_contexts : list of PresentationContext
        The Presentation Contexts supported by the local AE when acting
        as an Association *Acceptor*. Each item has values for Context ID,
        Abstract Syntax and Transfer Syntax.
    roles : dict or None
        If the *Requestor* has included one or more SCP/SCU Role Selection
        Negotiation items then this will be a :class:`dict` of
        ``{'SOP Class UID' : (SCU role, SCP role)}``, otherwise ``None``
        (default)

    Returns
    -------
    list of PresentationContext
        The accepted presentation context items, each with a Result value
        a Context ID, an Abstract Syntax and one Transfer Syntax item.
        Items are sorted in increasing Context ID value.
    list of SCP_SCU_RoleSelectionNegotiation
        If `roles` is not ``None`` then this is a :class:`list` of SCP/SCU Role
        Selection Negotiation items that can be sent back to the *Requestor*.
    """
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

    roles = roles or {}
    result_contexts: List[PresentationContext] = []
    reply_roles: Dict[UID, SCP_SCU_RoleSelectionNegotiation] = {}

    # No requestor presentation contexts
    if not rq_contexts:
        return result_contexts, []

    # Acceptor doesn't support any presentation contexts
    if not ac_contexts:
        for rq_context in rq_contexts:
            context = PresentationContext()
            context.context_id = rq_context.context_id
            context.abstract_syntax = rq_context.abstract_syntax
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            context.result = 0x03
            result_contexts.append(context)
        return result_contexts, []

    # Optimisation notes (for iterating through contexts only, not
    #   including actual context negotiation)
    # - Create dict, use set intersection/difference of dict keys: ~600 us
    # - Create dict, iterate over dict keys: ~400 us
    # - Iterate over lists: ~52000 us

    # Requestor may use the same Abstract Syntax in multiple Presentation
    #   Contexts so we need a more specific key than UID
    requestor_contexts = {(cx.context_id, cx.abstract_syntax): cx for cx in rq_contexts}
    # Acceptor supported SOP Classes must be unique so we can use UID as
    #   the key
    acceptor_contexts = {cx.abstract_syntax: cx for cx in ac_contexts}

    for (cntx_id, ab_syntax), rq_context in requestor_contexts.items():
        # Create a new PresentationContext item that will store the
        #   results of the negotiation
        context = PresentationContext()
        context.context_id = cntx_id
        context.abstract_syntax = ab_syntax
        context._as_scu = False
        context._as_scp = False

        # Check if the acceptor supports the Abstract Syntax
        if ab_syntax in acceptor_contexts:
            # Convenience variables
            ab_syntax = cast(UID, ab_syntax)
            ac_context = acceptor_contexts[ab_syntax]
            ac_roles = (ac_context.scu_role, ac_context.scp_role)
            rq_roles: Tuple[Optional[bool], Optional[bool]]
            try:
                rq_roles = roles[ab_syntax]
                has_role = True
            except KeyError:
                rq_roles = (None, None)
                has_role = False

            # Abstract syntax supported so check Transfer Syntax
            for tr_syntax in ac_context.transfer_syntax:
                # If transfer syntax supported then (provisionally) accept
                if tr_syntax in rq_context.transfer_syntax:
                    context.transfer_syntax = [tr_syntax]
                    context.result = 0x00
                    result_contexts.append(context)
                    break

            # SCP/SCU Role Selection Negotiation
            #   Only for (provisionally) accepted contexts
            if context.result == 0x00:
                if None in ac_roles:
                    # Default roles
                    context._as_scu = False
                    context._as_scp = True
                    # If either ac.scu_role or ac.scp_role is None then
                    #   don't send an SCP/SCU negotiation reply
                    has_role = False
                else:
                    # Use a LUT to make changes to outcomes easier
                    #   also its much simpler than coding if/then branches
                    outcome = SCP_SCU_ROLES[rq_roles][ac_roles]
                    context._as_scu = outcome[2]
                    context._as_scp = outcome[3]

                # If can't act as either SCU nor SCP then reject the context
                if context.as_scu is False and context.as_scp is False:
                    # User rejection
                    context.result = 0x01

            # Need to check against None as 0x00 is a possible value
            if context.result is None:
                # Reject context - transfer syntax not supported
                context.result = 0x04
                context.transfer_syntax = [rq_context.transfer_syntax[0]]
                result_contexts.append(context)
            elif context.result == 0x00 and has_role:
                # Create new SCP/SCU Role Selection Negotiation item
                role = SCP_SCU_RoleSelectionNegotiation()
                role.sop_class_uid = context.abstract_syntax

                # Can't return 0x01 if proposed 0x00
                if rq_roles[0] is False:
                    role.scu_role = False
                else:
                    role.scu_role = ac_context.scu_role

                if rq_roles[1] is False:
                    role.scp_role = False
                else:
                    role.scp_role = ac_context.scp_role

                reply_roles[cast(UID, context.abstract_syntax)] = role
        else:
            # Reject context - abstract syntax not supported
            context.result = 0x03
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            result_contexts.append(context)

    # Sort by presentation context ID
    #   This isn't required by the DICOM Standard but its a nice thing to do
    result_contexts = sorted(result_contexts, key=lambda x: cast(int, x.context_id))

    # Sort role selection by abstract syntax, also not required but nice
    result_roles = sorted(
        reply_roles.values(), key=lambda x: cast(UID, x.sop_class_uid)
    )

    return result_contexts, result_roles


def negotiate_as_requestor(
    rq_contexts: ListCXType, ac_contexts: ListCXType, roles: RoleType = None
) -> ListCXType:
    """Process the Presentation Contexts as an Association *Requestor*.

    The *Acceptor* has processed the *Requestor's* presentation context
    definition list and returned the results. We want to do two things:

    - Process the SCP/SCU Role Selection Negotiation items
    - Return a nice list of :class:`PresentationContext` with the Results and
      original Abstract Syntax values to make things easier to use.

    :class:`~pynetdicom.pdu_items.PresentationContextItemRQ`

    - Presentation context ID
    - Abstract Syntax: one
    - Transfer syntax: one or more

    :class:`~pynetdicom.pdu_items.PresentationContextItemAC`

    - Presentation context ID
    - Result: one of ``0x00``, ``0x01``, ``0x02``, ``0x03`` or ``0x04``
    - Transfer syntax: one, not to be tested if result is not ``0x00``

    Parameters
    ----------
    rq_contexts : list of PresentationContext
        The Presentation Contexts sent to the peer as the A-ASSOCIATE's
        Presentation Context Definition List.
    ac_contexts : list of PresentationContext
        The Presentation Contexts return by the peer as the A-ASSOCIATE's
        Presentation Context Definition Result List.
    roles : dict or None
        If the *Acceptor* has included one or more SCP/SCU Role Selection
        Negotiation items then this will be a :class:`dict` of
        ``{'SOP Class UID' : (SCU role, SCP role)}``, otherwise ``None``
        (default)

    Returns
    -------
    list of PresentationContext
        The contexts in the returned Presentation Context Definition Result
        List, with added Abstract Syntax value. Items are sorted in
        increasing Context ID value and the SCP/SCU roles are set as per
        the negotiated outcome.
    """
    roles = roles or {}

    if not rq_contexts:
        raise ValueError("Requestor contexts are required")
    output = []

    # Create dicts, indexed by the presentation context ID
    requestor_contexts = {context.context_id: context for context in rq_contexts}
    acceptor_contexts = {context.context_id: context for context in ac_contexts}

    for context_id in requestor_contexts:
        # Convenience variable
        rq_context = requestor_contexts[context_id]

        context = PresentationContext()
        context.context_id = context_id
        context.abstract_syntax = rq_context.abstract_syntax
        # Ensure we always have a role set
        context._as_scu = False
        context._as_scp = False

        if context_id in acceptor_contexts:
            # Convenience variable
            ac_context = acceptor_contexts[context_id]

            # Update with accepted values
            if ac_context.transfer_syntax:
                context.transfer_syntax = [ac_context.transfer_syntax[0]]
            context.result = ac_context.result

            # SCP/SCU Role Selection Negotiation
            rq_roles = (rq_context.scu_role, rq_context.scp_role)
            ac_roles: Tuple[Optional[bool], Optional[bool]]
            try:
                ac_roles = roles[cast(UID, context.abstract_syntax)]
            except KeyError:
                ac_roles = (None, None)

            # Skip if context rejected or acceptor ignored proposal
            if ac_context.result == 0x00 and None not in ac_roles:
                outcome = SCP_SCU_ROLES[rq_roles][ac_roles]
                context._as_scu = outcome[0]
                context._as_scp = outcome[1]
            else:
                # We are the association requestor, so SCU role only
                context._as_scp = False
                context._as_scu = True

        # Add any missing contexts as rejected
        else:
            context.transfer_syntax = [rq_context.transfer_syntax[0]]
            context.result = 0x02

        output.append(context)

    # Sort returned list by context ID
    return sorted(output, key=lambda x: cast(int, x.context_id))


def build_context(
    abstract_syntax: Union[str, UID],
    transfer_syntax: Optional[Union[str, UID, List[Union[str, UID]]]] = None,
) -> PresentationContext:
    """Return a :class:`PresentationContext` built from the `abstract_syntax`.

    Parameters
    ----------
    abstract_syntax : str or UID or sop_class.SOPClass
        The :class:`~pydicom.uid.UID` or subclass of
        :class:`~pynetdicom.sop_class.SOPClass` instance to use as the abstract
        syntax.
    transfer_syntax : str/UID or list of str/UID
        The transfer syntax UID(s) to use (default:
        ``[Implicit VR Little Endian, Explicit VR Little Endian,
        Deflated Explicit VR Little Endian, Implicit VR Big Endian]``)

    Examples
    --------

    Specifying a presentation context with the default transfer syntaxes

    >>> from pynetdicom import build_context
    >>> context = build_context('1.2.840.10008.1.1')
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian
        =Explicit VR Little Endian
        =Deflated Explicit VR Little Endian
        =Explicit VR Big Endian

    Specifying the abstract syntax using a *pynetdicom*
    :class:`~pynetdicom.sop_class.SOPClass` instance and a single transfer
    syntax

    >>> from pynetdicom import build_context
    >>> from pynetdicom.sop_class import Verification
    >>> context = build_context(Verification, '1.2.840.10008.1.2')
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian

    Specifying multiple transfer syntaxes

    >>> from pydicom.uid import UID
    >>> from pynetdicom import build_context
    >>> context = build_context(
    ...     UID('1.2.840.10008.1.1'), ['1.2.840.10008.1.2', '1.2.840.10008.1.2.4.50']
    ... )
    >>> print(context)
    Abstract Syntax: Verification SOP Class
    Transfer Syntax(es):
        =Implicit VR Little Endian
        =JPEG Baseline (Process 1)

    Returns
    -------
    presentation.PresentationContext
    """
    if transfer_syntax is None:
        transfer_syntax = DEFAULT_TRANSFER_SYNTAXES

    abstract_syntax = UID(abstract_syntax)

    # Allow single transfer syntax values for convenience
    if isinstance(transfer_syntax, str):
        transfer_syntax = [transfer_syntax]

    context = PresentationContext()
    context.abstract_syntax = abstract_syntax
    context.transfer_syntax = transfer_syntax  # type: ignore

    return context


def build_role(
    uid: Union[str, UID], scu_role: bool = False, scp_role: bool = False
) -> "SCP_SCU_RoleSelectionNegotiation":
    """Return a SCP/SCU Role Selection Negotiation item.

    .. versionadded:: 1.2

    Parameters
    ----------
    uid : str or UID or sop_class.SOPClass
        The :class:`~pydicom.uid.UID` or subclass of
        :class:`~pynetdicom.sop_class.SOPClass` instance to use as the *SOP
        Class UID* parameter value.
    scu_role : bool, optional
        ``True`` to propose the SCU role for the *Requestor*, ``False``
        otherwise (default).
    scp_role : bool, optional
        ``True`` to propose the SCP role for the *Requestor*, ``False``
        otherwise (default).

    Returns
    -------
    pdu_primitives.SCP_SCU_RoleSelectionNegotiation
        The role selection item.
    """
    from pynetdicom.pdu_primitives import SCP_SCU_RoleSelectionNegotiation

    role = SCP_SCU_RoleSelectionNegotiation()
    role.sop_class_uid = UID(uid)
    role.scu_role = scu_role
    role.scp_role = scp_role
    return role


# Service specific pre-generated Presentation Contexts
# pylint: disable=line-too-long,invalid-name
ApplicationEventLoggingPresentationContexts = [
    build_context(uid) for uid in sorted(_APPLICATION_EVENT_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Application Event Logging<part04/chapter_P.html>`."""

BasicWorklistManagementPresentationContexts = [
    build_context(uid) for uid in sorted(_BASIC_WORKLIST_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Basic Worklist Management<part04/chapter_K.html>`."""

ColorPalettePresentationContexts = [
    build_context(uid) for uid in sorted(_COLOR_PALETTE_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Color Palette Query/Retrieve<part04/chapter_X.html>`."""

DefinedProcedureProtocolPresentationContexts = [
    build_context(uid) for uid in sorted(_DEFINED_PROCEDURE_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Defined Procedure Protocol Query/Retrieve<part04/chapter_HH.html>`."""

DisplaySystemPresentationContexts = [
    build_context(uid) for uid in sorted(_DISPLAY_SYSTEM_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Display System Management<part04/chapter_EE.html>`."""

HangingProtocolPresentationContexts = [
    build_context(uid) for uid in sorted(_HANGING_PROTOCOL_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Hanging Protocol Query/Retrieve<part04/chapter_U.html>`."""

ImplantTemplatePresentationContexts = [
    build_context(uid) for uid in sorted(_IMPLANT_TEMPLATE_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Implant Template Query/Retrieve<part04/chapter_BB.html>`."""

InstanceAvailabilityPresentationContexts = [
    build_context(uid) for uid in sorted(_INSTANCE_AVAILABILITY_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Instance Availability Notification<part04/chapter_R.html>`."""

MediaCreationManagementPresentationContexts = [
    build_context(uid) for uid in sorted(_MEDIA_CREATION_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Media Creation Management<part04/chapter_S.html>`."""

MediaStoragePresentationContexts = [
    build_context(uid) for uid in sorted(_MEDIA_STORAGE_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Media Storage<part04/chapter_I.html>`."""

ModalityPerformedPresentationContexts = [
    build_context(uid) for uid in sorted(_PROCEDURE_STEP_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Modality Performed Procedure Step<part04/chapter_F.html>`."""

NonPatientObjectPresentationContexts = [
    build_context(uid) for uid in sorted(_NON_PATIENT_OBJECT_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Non-Patient Object Storage<part04/chapter_GG.html>`."""

PrintManagementPresentationContexts = [
    build_context(uid) for uid in sorted(_PRINT_MANAGEMENT_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Print Management<part04/chapter_H.html>`."""

ProcedureStepPresentationContexts = [
    build_context(uid) for uid in sorted(_PROCEDURE_STEP_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Procedure Step<part04/chapter_F.html>`."""

ProtocolApprovalPresentationContexts = [
    build_context(uid) for uid in sorted(_PROTOCOL_APPROVAL_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Protocol Approval Query/Retrieve<part04/chapter_II.html>`."""

QueryRetrievePresentationContexts = [
    build_context(uid) for uid in sorted(_QR_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Query/Retrieve<part04/chapter_C.html>`."""

RelevantPatientInformationPresentationContexts = [
    build_context(uid) for uid in sorted(_RELEVANT_PATIENT_QUERY_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Relevant Patient Information Query<part04/chapter_Q.html>`."""

RTMachineVerificationPresentationContexts = [
    build_context(uid) for uid in sorted(_RT_MACHINE_VERIFICATION_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`RT Machine Verification<part04/chapter_DD.html>`."""

AllStoragePresentationContexts = [
    build_context(uid) for uid in sorted(_STORAGE_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Storage<part04/chapter_B.html>` containing all SOP Classes."""

StoragePresentationContexts = AllStoragePresentationContexts[:128]
"""Pre-built presentation contexts for :dcm:`Storage<part04/chapter_B.html>` containing the first 128 SOP Classes."""

StorageCommitmentPresentationContexts = [
    build_context(uid) for uid in sorted(_STORAGE_COMMITMENT_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Storage Commitment<part04/chapter_J.html>`."""

SubstanceAdministrationPresentationContexts = [
    build_context(uid) for uid in sorted(_SUBSTANCE_ADMINISTRATION_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Substance Administration Query<part04/chapter_V.html>`."""

UnifiedProcedurePresentationContexts = [
    build_context(uid) for uid in sorted(_UNIFIED_PROCEDURE_STEP_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Unified Procedure Step<part04/chapter_CC.html>`."""

VerificationPresentationContexts = [
    build_context(uid) for uid in sorted(_VERIFICATION_CLASSES.values())
]
"""Pre-built presentation contexts for :dcm:`Verification<part04/chapter_A.html>`."""



from setuptools import setup, find_packages
from pathlib import Path


BASE_DIR = Path(__file__).parent

VERSION_FILE = BASE_DIR / "pynetdicom" / "_version.py"
with open(VERSION_FILE) as f:
    exec(f.read())

with open(BASE_DIR / "README.rst", "r") as f:
    long_description = f.read()

setup(
    name="pynetdicom",
    packages=find_packages(),
    include_package_data=True,
    version=__version__,
    zip_safe=False,
    description="A Python implementation of the DICOM networking protocol",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="",
    author_email="scaramallion@users.noreply.github.com",
    url="https://github.com/pydicom/pynetdicom",
    license="MIT",
    keywords=(
        "dicom network python medicalimaging radiotherapy oncology pydicom imaging"
    ),
    project_urls={"Documentation": "https://pydicom.github.io/pynetdicom/"},
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Development Status :: 5 - Production/Stable",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Medical Science Apps.",
        "Topic :: Software Development :: Libraries",
    ],
    install_requires=["pydicom>=2.2.0"],
    extras_require={  # will also install from `install_requires`
        "apps": ["sqlalchemy"],
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
            "sphinx-copybutton",
            "sphinx-issues",
            "sphinxcontrib-napoleon",
            "numpydoc",
        ],
        "tests": ["pytest", "pyfakefs", "sqlalchemy"],
    },
    python_requires=">=3.7",
)



"""DICOM Upper Layer Protocol Data Units (PDUs).

There are seven different PDUs:

- A_ASSOCIATE_RQ
- A_ASSOCIATE_AC
- A_ASSOCIATE_RJ
- P_DATA_TF
- A_RELEASE_RQ
- A_RELEASE_RP
- A_ABORT_RQ

::

                  from_primitive               encode
  +----------------+  ------>  +------------+  ----->  +-------------+
  | DUL Primitive  |           |    PDU     |          |   Peer AE   |
  +----------------+  <------  +------------+  <-----  +-------------+
                    to_primitive               decode
"""

import logging
from struct import Struct
from typing import (
    Iterator,
    Optional,
    Any,
    Callable,
    List,
    Tuple,
    TYPE_CHECKING,
    Union,
    cast,
)

from pydicom.uid import UID

from pynetdicom.pdu_items import (
    ApplicationContextItem,
    PresentationContextItemRQ,
    PresentationContextItemAC,
    UserInformationItem,
    PresentationDataValueItem,
    PDU_ITEM_TYPES,
    _PDUItemType,
    PDUItem,
)
from pynetdicom.utils import decode_bytes, set_ae

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.pdu_primitives import (
        A_ASSOCIATE,
        P_DATA,
        A_RELEASE,
        A_ABORT,
        A_P_ABORT,
    )


LOGGER = logging.getLogger("pynetdicom.pdu")

_PDVItem = List[PresentationDataValueItem]
_AbortType = Union["A_ABORT", "A_P_ABORT"]
_PDUType = Union[
    "A_ASSOCIATE_RQ",
    "A_ASSOCIATE_AC",
    "A_ASSOCIATE_RJ",
    "P_DATA_TF",
    "A_RELEASE_RQ",
    "A_RELEASE_RP",
    "A_ABORT_RQ",
]

# Predefine some structs to make decoding and encoding faster
UCHAR = Struct("B")
UINT2 = Struct(">H")
UINT4 = Struct(">I")

UNPACK_UCHAR = UCHAR.unpack
UNPACK_UINT2 = UINT2.unpack
UNPACK_UINT4 = UINT4.unpack

PACK_UCHAR = UCHAR.pack
PACK_UINT2 = UINT2.pack
PACK_UINT4 = UINT4.pack


class PDU:
    """Base class for PDUs.

    Protocol Data Units (PDUs) are the message formats exchanged between peer
    entities within a layer. A PDU consists of protocol control information
    and user data. PDUs are constructed by mandatory fixed fields followed by
    optional variable fields that contain one or more items and/or sub-items.

    References
    ----------
    DICOM Standard, Part 8, :dcm:`Section 9.3 <part08/sect_9.3.html>`
    """

    def decode(self, bytestream: bytes) -> None:
        """Decode `bytestream` and use the result to set the field values of
        the PDU.

        Parameters
        ----------
        bytestream : bytes
            The PDU data to be decoded.
        """
        for (offset, length), attr_name, func, args in self._decoders:
            # Allow us to use None as a `length`
            if length:
                sl = slice(offset, offset + length)
            else:
                sl = slice(offset, None)

            setattr(self, attr_name, func(bytestream[sl], *args))

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders."""
        raise NotImplementedError

    def encode(self) -> bytes:
        """Return the encoded PDU as :class:`bytes`.

        Returns
        -------
        bytes
            The encoded PDU.
        """
        bytestream = bytes()
        for attr_name, func, args in self._encoders:
            # If attr_name is None then the field is usually reserved
            if attr_name:
                bytestream += func(getattr(self, attr_name), *args)
            else:
                bytestream += func(*args)

        return bytestream

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field encoders."""
        raise NotImplementedError

    def __eq__(self, other: Any) -> bool:
        """Return ``True`` if `self` equals `other`."""
        if other is self:
            return True

        # pylint: disable=protected-access
        if isinstance(other, self.__class__):
            self_dict = {
                enc[0]: getattr(self, enc[0]) for enc in self._encoders if enc[0]
            }
            other_dict = {
                enc[0]: getattr(other, enc[0]) for enc in other._encoders if enc[0]
            }
            return self_dict == other_dict

        return NotImplemented

    @staticmethod
    def _generate_items(bytestream: bytes) -> Iterator[Tuple[int, bytes]]:
        """Yield PDU item data from `bytestream`.

        Parameters
        ----------
        bytestream : bytes
            The encoded PDU variable item data.

        Yields
        ------
        int, bytes
            The variable item's 'Item Type' parameter as int, and the item's
            entire encoded data as bytes.

        Notes
        -----
        Can be used with the following PDU items/sub-items:

        - Application Context Item
        - Presentation Context Item (RQ/AC)

          - Abstract Syntax Sub-item
          - Transfer Syntax Sub-item
        - User Information Item

          - Implementation Class UID Sub-item (RQ/AC)
          - Implementation Version Name Sub-item (RQ/AC)
          - Asynchronous Operations Window Sub-item (RQ/AC)
          - SCP/SCU Role Selection Sub-item (RQ/AC)
          - SOP Class Extended Negotiation Sub-item (RQ/AC)
          - SOP Class Common Extended Negotiation Sub-item (RQ/AC)
          - User Identity Sub-item (RQ/AC)

        **Encoding**
        When encoded, PDU item and sub-item data for the above has the
        following structure, taken from various tables in (offsets shown
        with Python indexing). Items are always encoded using Big Endian.

        +--------+-------------+-------------+
        | Offset | Length      | Description |
        +========+=============+=============+
        | 0      | 1           | Item type   |
        +--------+-------------+-------------+
        | 1      | 1           | Reserved    |
        +--------+-------------+-------------+
        | 2      | 2           | Item length |
        +--------+-------------+-------------+
        | 4      | Item length | Item data   |
        +--------+-------------+-------------+

        References
        ----------
        * DICOM Standard, Part 8, :dcm:`Section 9.3 <part08/sect_9.3.html>`
        * DICOM Standard, Part 8,
           :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
        """
        offset = 0
        while bytestream[offset : offset + 1]:
            item_type = UNPACK_UCHAR(bytestream[offset : offset + 1])[0]
            item_length = UNPACK_UINT2(bytestream[offset + 2 : offset + 4])[0]
            item_data = bytestream[offset : offset + 4 + item_length]
            assert len(item_data) == 4 + item_length
            yield item_type, item_data
            # Move `offset` to the start of the next item
            offset += 4 + item_length

    def __len__(self) -> int:
        """Return the total length of the encoded PDU as :class:`int`."""
        return 6 + self.pdu_length

    def __ne__(self, other: Any) -> bool:
        """Return ``True`` if `self` does not equal `other`."""
        return not self == other

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as :class:`int`."""
        raise NotImplementedError

    @property
    def pdu_type(self) -> int:
        """Return the *PDU Type* field value as :class:`int`."""
        return PDU_TYPES[self.__class__]

    @staticmethod
    def _wrap_bytes(bytestream: bytes) -> bytes:
        """Return `bytestream` without changing it."""
        return bytestream

    @staticmethod
    def _wrap_encode_items(items: List[PDUItem]) -> bytes:
        """Return `items` encoded as bytes.

        Parameters
        ----------
        items : list of PDU items
            The items to encode.

        Returns
        -------
        bytes
            The encoded items.
        """
        bytestream = bytes()
        for item in items:
            bytestream += item.encode()

        return bytestream

    @staticmethod
    def _wrap_encode_str(value: str, pad: int = 0) -> bytes:
        """Return `value` as ASCII encoded :class:`bytes`.

         Each component of Application Context, Abstract Syntax and Transfer
         Syntax UIDs should be encoded as a ISO 646:1990-Basic G0 Set Numeric
         String (characters 0-9), with each component separated by '.' (0x2e)
        .

         'ascii' is chosen because this is the codec Python uses for ISO 646
         [3]_.

         Parameters
         ----------
         value : str
             The ASCII string to be encoded.
         pad : int, optional
             The amount of trailing padding to be used (default ``0``).

         Returns
         -------
         bytes
             The encoded `value`.

         References
         ----------
         * DICOM Standard, Part 8, :dcm:`Annex F <part08/chapter_F.html>`
         * `Python 3 codecs module
           <https://docs.python.org/2/library/codecs.html#standard-encodings>`_
        """
        return value.ljust(pad).encode("ascii", errors="strict")

    def _wrap_generate_items(self, bytestream: bytes) -> List[PDUItem]:
        """Return a list of decoded PDU items generated from `bytestream`."""
        item_list = []
        for item_type, item_bytes in self._generate_items(bytestream):
            item = PDU_ITEM_TYPES[item_type]()
            item.decode(item_bytes)
            item_list.append(item)

        return item_list

    @staticmethod
    def _wrap_pack(value: Any, packer: Callable[[Any], bytes]) -> bytes:
        """Return `value` encoded as bytes using `packer`.

        Parameters
        ----------
        value
            The value to encode.
        packer : callable
            A callable function to use to pack the data as bytes. The
            `packer` should return the packed bytes. Example:
            struct.Struct('>I').pack

        Returns
        -------
        bytes
        """
        return packer(value)

    @staticmethod
    def _wrap_unpack(bytestream: bytes, unpacker: Callable[[bytes], Tuple[Any]]) -> Any:
        """Return the first value when `unpacker` is run on `bytestream`.

        Parameters
        ----------
        bytestream : bytes
            The encoded data to unpack.
        unpacker : callable
            A callable function to use to unpack the data in `bytestream`. The
            `unpacker` should return a tuple containing unpacked values.
            Example: struct.Struct('>I').unpack.
        """
        return unpacker(bytestream)[0]


class A_ASSOCIATE_RQ(PDU):
    """An A-ASSOCIATE-RQ PDU.

    An A-ASSOCIATE-RQ PDU is sent by an association requestor to initiate
    association negotiation with an acceptor.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x01``).
    protocol_version : int
        The *Protocol Version* field value (``0x01``).
    variable_items : list
        A list containing the A-ASSOCIATE-RQ's *Variable Items*. Contains
        one Application Context item, one or more Presentation Context items
        and one User Information item. The order of the items is not
        guaranteed.

    Notes
    -----
    An A-ASSOCIATE-RQ PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x01``)
    * PDU length (1)
    * Protocol version (1, default value, ``0x01``)
    * Called AE title (1)
    * Calling AE title (1)
    * Variable items (1)

      * Application Context Item (1)

        * Item type (1, fixed value, ``0x10``)
        * Item length (1)
        * Application Context Name (1, fixed in an application)
      * Presentation Context Item(s) (1 or more)

        * Item type (1, fixed value, ``0x21``)
        * Item length (1)
        * Context ID (1)
        * Abstract/Transfer Syntax Sub-items (1)

          * Abstract Syntax Sub-item (1)

            * Item type (1, fixed, ``0x30``)
            * Item length (1)
            * Abstract syntax name (1)
          * Transfer Syntax Sub-items (1 or more)

            * Item type (1, fixed, ``0x40``)
            * Item length (1)
            * Transfer syntax name(s) (1 or more)
      * User Information Item (1)

        * Item type (1, fixed, ``0x50``)
        * Item length (1)
        * User data Sub-items (2 or more)

            * Maximum Length Received Sub-item (1)
            * Implementation Class UID Sub-item (1)
            * Optional User Data Sub-items (0 or more)

    **Encoding**

    When encoded, an A-ASSOCIATE-RQ PDU has the following structure, taken
    from `Table 9-11<part08/sect_9.3.2.html>` (offsets shown with Python
    indexing). PDUs are always encoded using Big Endian.

    +--------+-------------+------------------+
    | Offset | Length      | Description      |
    +========+=============+==================+
    | 0      | 1           | PDU type         |
    +--------+-------------+------------------+
    | 1      | 1           | Reserved         |
    +--------+-------------+------------------+
    | 2      | 4           | PDU length       |
    +--------+-------------+------------------+
    | 6      | 2           | Protocol version |
    +--------+-------------+------------------+
    | 8      | 2           | Reserved         |
    +--------+-------------+------------------+
    | 10     | 16          | Called AE title  |
    +--------+-------------+------------------+
    | 26     | 16          | Calling AE title |
    +--------+-------------+------------------+
    | 42     | 32          | Reserved         |
    +--------+-------------+------------------+
    | 74     | Variable    | Variable items   |
    +--------+-------------+------------------+

    References
    ----------
    * DICOM Standard, Part 8, Sections :dcm:`9.3.2<part08/sect_9.3.2.html>`
      and :dcm:`9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["A_ASSOCIATE"] = None) -> None:
        """Initialise a new A-ASSOCIATE-RQ PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_ASSOCIATE
            The primitive to use to initialise the PDU.
        """
        # We allow the user to modify the protocol version if so desired
        self.protocol_version = 0x01
        self._called_aet = ""
        self._calling_aet = ""

        # Set some default values
        self.called_ae_title = "Default"
        self.calling_ae_title = "Default"

        # `variable_items` is a list containing the following:
        #   1 ApplicationContextItem
        #   1 or more PresentationContextItemRQ
        #   1 UserInformationItem
        # The order of the items in the list may not be as given above
        self.variable_items: _PDUItemType = []

        if primitive is not None:
            self.from_primitive(primitive)

    def from_primitive(self, primitive: "A_ASSOCIATE") -> None:
        """Setup the current PDU using an A-ASSOCIATE (request) primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_ASSOCIATE
            The primitive to use to set the current PDU field values.
        """
        self.calling_ae_title = primitive.calling_ae_title
        self.called_ae_title = primitive.called_ae_title

        # Add Application Context
        application_context = ApplicationContextItem()
        application_context.application_context_name = cast(
            UID, primitive.application_context_name
        )
        self.variable_items.append(application_context)

        # Add Presentation Context(s)
        for contexts in primitive.presentation_context_definition_list:
            presentation_context = PresentationContextItemRQ()
            presentation_context.from_primitive(contexts)
            self.variable_items.append(presentation_context)

        # Add User Information
        user_information = UserInformationItem()
        user_information.from_primitive(primitive.user_information)
        self.variable_items.append(user_information)

    def to_primitive(self) -> "A_ASSOCIATE":
        """Return an A-ASSOCIATE (request) primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_ASSOCIATE
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_ASSOCIATE

        primitive = A_ASSOCIATE()
        primitive.calling_ae_title = self.calling_ae_title
        primitive.called_ae_title = self.called_ae_title
        primitive.application_context_name = self.application_context_name

        for item in self.variable_items:
            # Add presentation contexts
            if isinstance(item, PresentationContextItemRQ):
                primitive.presentation_context_definition_list.append(
                    item.to_primitive()
                )

            # Add user information
            elif isinstance(item, UserInformationItem):
                primitive.user_information = item.to_primitive()

        return primitive

    @property
    def application_context_name(self) -> Optional[UID]:
        """Return the *Application Context Name*, if available.

        Returns
        -------
        pydicom.uid.UID or None
            The requestor's *Application Context Name* or None if not
            available.
        """
        for item in self.variable_items:
            if isinstance(item, ApplicationContextItem):
                return item.application_context_name

        return None

    @property
    def called_ae_title(self) -> str:
        """Get or set the *Called AE Title* field value as :class:`str`.

        Will be converted to a fixed length 16-byte value (padded with trailing
        spaces ``0x20``). Leading and trailing spaces are non-significant and a
        value of 16 spaces is not allowed.

        Parameters
        ----------
        value : str or bytes
            The value you wish to set. A value consisting of spaces is not
            allowed and values longer than 16 characters will be truncated.
        """
        return self._called_aet

    @called_ae_title.setter
    def called_ae_title(self, value: Union[str, bytes]) -> None:
        """Set the *Called AE Title* field value"""
        if isinstance(value, bytes):
            # PS3.8 Table 9-11: Leading and trailing spaces are non-significant
            value = decode_bytes(value).strip()
            if not value:
                raise ValueError(
                    "Invalid 'Called AE Title' value - must not consist "
                    "entirely of spaces"
                )

        self._called_aet = cast(str, set_ae(value, "Called AE Title", False, False))

    @property
    def calling_ae_title(self) -> str:
        """Return the *Calling AE Title* field value as :class:`str`."""
        return self._calling_aet

    @calling_ae_title.setter
    def calling_ae_title(self, value: Union[str, bytes]) -> None:
        """Set the *Calling AE Title* field value.

        Will be converted to a fixed length 16-byte value (padded with trailing
        spaces ``0x20``). Leading and trailing spaces are non-significant and a
        value of 16 spaces is not allowed.

        Parameters
        ----------
        value : str or bytes
            The value you wish to set. A value consisting of spaces is not
            allowed and values longer than 16 characters will be truncated.
        """
        if isinstance(value, bytes):
            # PS3.8 Table 9-11: Leading and trailing spaces are non-significant
            value = decode_bytes(value).strip()
            if not value:
                raise ValueError(
                    "Invalid 'Calling AE Title' value - must not consist "
                    "entirely of spaces"
                )

        self._calling_aet = cast(str, set_ae(value, "Calling AE Title", False, False))

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ``((offset, length), attr_name, callable, [args])``,
            where:

            - ``offset`` is the byte offset to start at
            - ``length`` is how many bytes to slice (if None then will slice
              to the end of the data),
            - ``attr_name`` is the name of the attribute corresponding to the
              field
            - ``callable`` is a decoding function that returns the decoded
              value
            - ``args`` is a list of arguments to pass ``callable``
        """
        return [
            ((6, 2), "protocol_version", self._wrap_unpack, [UNPACK_UINT2]),
            ((10, 16), "called_ae_title", self._wrap_bytes, []),
            ((26, 16), "calling_ae_title", self._wrap_bytes, []),
            ((74, None), "variable_items", self._wrap_generate_items, []),
        ]

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ``(attr_name, callable, [args])``, where:

            - ``attr_name`` is the name of the attribute corresponding to the
              field
            - ``callable`` is an encoding function that returns :class:`bytes`
            - ``args`` is a :class:`list` of arguments to pass ``callable``.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            ("protocol_version", PACK_UINT2, []),
            (None, self._wrap_pack, [0x0000, PACK_UINT2]),
            ("called_ae_title", self._wrap_encode_str, [16]),
            ("calling_ae_title", self._wrap_encode_str, [16]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            ("variable_items", self._wrap_encode_items, []),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as :class:`int`."""
        length = 68
        for item in self.variable_items:
            length += len(item)

        return length

    @property
    def presentation_context(self) -> List[PresentationContextItemRQ]:
        """Return a list of the Presentation Context items.

        Returns
        -------
        list of pdu_items.PresentationContextItemRQ
            The Presentation Context items.
        """
        return [
            item
            for item in self.variable_items
            if isinstance(item, PresentationContextItemRQ)
        ]

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = ["A-ASSOCIATE-RQ PDU"]
        s.append("==================")
        s.append(f"  PDU type: 0x{self.pdu_type:02X}")
        s.append(f"  PDU length: {self.pdu_length} bytes")
        s.append(f"  Protocol version: {self.protocol_version}")
        s.append(f"  Called AET:  {self.called_ae_title}")
        s.append(f"  Calling AET: {self.calling_ae_title}")
        s.append("")

        s.append("  Variable Items")
        s.append("  ---------------")
        s.append("  * Application Context Item")
        s.append(f"    - Context name: ={self.application_context_name}")
        s.append("  * Presentation Context Item(s):")

        for cx in self.presentation_context:
            item_str_list = str(cx).split("\n")
            s.append(f"    - {item_str_list[0]}")
            for jj in item_str_list[1:-1]:
                s.append(f"      {jj}")

        s.append("  * User Information Item(s):")
        for item in cast(UserInformationItem, self.user_information).user_data:
            item_str_list = str(item).split("\n")
            s.append(f"    - {item_str_list[0]}")
            for jj in item_str_list[1:-1]:
                s.append(f"      {jj}")

        return "\n".join(s)

    @property
    def user_information(self) -> Optional[UserInformationItem]:
        """Return the User Information Item, if available.

        Returns
        -------
        pdu_items.UserInformationItem or None
            The requestor's User Information object or ``None``, if not
            available.
        """
        for item in self.variable_items:
            if isinstance(item, UserInformationItem):
                return item

        return None


class A_ASSOCIATE_AC(PDU):
    """An A-ASSOCIATE-AC PDU.

    An A-ASSOCIATE-AC PDU is sent by an association acceptor to indicate that
    association negotiation has been successful.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x02``).
    protocol_version : int
        The *Protocol Version* field value (default ``0x01``).
    variable_items : list
        A list containing the A-ASSOCIATE-AC's 'Variable Items'. Contains
        one Application Context item, one or more Presentation Context items
        and one User Information item. The order of the items is not
        guaranteed.

    Notes
    -----
    An A-ASSOCIATE-AC PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x02``)
    * PDU length (1)
    * Protocol version (1, default value, ``0x01``)
    * Variable items (1)

      * Application Context Item (1)

        * Item type (1, fixed value, ``0x10``)
        * Item length (1)
        * Application Context Name (1, fixed in an application)
      * Presentation Context Item(s) (1 or more)

        * Item type (1, fixed value, ``0x21``)
        * Item length (1)
        * Context ID (1)
        * Result/reason (1)
        * Transfer Syntax Sub-items (1)

          * Item type (1, fixed, ``0x40``)
          * Item length (1)
          * Transfer syntax name(s) (1)
      * User Information Item (1)

        * Item type (1, fixed, ``0x50``)
        * Item length (1)
        * User data Sub-items (2 or more)

            * Maximum Length Received Sub-item (1)
            * Implementation Class UID Sub-item (1)
            * Optional User Data Sub-items (0 or more)

    **Encoding**

    When encoded, an A-ASSOCIATE-AC PDU has the following structure, taken
    from Table 9-17 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+------------------+
    | Offset | Length      | Description      |
    +========+=============+==================+
    | 0      | 1           | PDU type         |
    +--------+-------------+------------------+
    | 1      | 1           | Reserved         |
    +--------+-------------+------------------+
    | 2      | 4           | PDU length       |
    +--------+-------------+------------------+
    | 6      | 2           | Protocol version |
    +--------+-------------+------------------+
    | 8      | 2           | Reserved         |
    +--------+-------------+------------------+
    | 10     | 16          | Reserved^        |
    +--------+-------------+------------------+
    | 26     | 16          | Reserved^        |
    +--------+-------------+------------------+
    | 42     | 32          | Reserved         |
    +--------+-------------+------------------+
    | 74     | Variable    | Variable items   |
    +--------+-------------+------------------+

    ^ The reserved fields shall be sent with a value identical to the value
    received in the A-ASSOCIATE-RQ but their values shall not be tested.

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.3 <part08/sect_9.3.3.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["A_ASSOCIATE"] = None) -> None:
        """Initialise a new A-ASSOCIATE-AC PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_ASSOCIATE
            The primitive to use to initialise the PDU.
        """
        # We allow the user to modify the protocol version if so desired
        self.protocol_version = 0x01
        # Called AE Title, should be present, but no guarantees
        self._reserved_aet: str = ""
        # Calling AE Title, should be present, but no guarantees
        self._reserved_aec: str = ""

        # `variable_items` is a list containing the following:
        #   1 ApplicationContextItem
        #   1 or more PresentationContextItemAC
        #   1 UserInformationItem
        # The order of the items in the list may not be as given above
        self.variable_items: _PDUItemType = []

        if primitive is not None:
            self.from_primitive(primitive)

    def from_primitive(self, primitive: "A_ASSOCIATE") -> None:
        """Setup the current PDU using an A-ASSOCIATE (accept) primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_ASSOCIATE
            The primitive to use to set the current PDU field values.
        """
        self.reserved_aet = primitive.called_ae_title
        self.reserved_aec = primitive.calling_ae_title

        # Make application context
        application_context = ApplicationContextItem()
        application_context.application_context_name = cast(
            UID, primitive.application_context_name
        )
        self.variable_items.append(application_context)

        # Make presentation contexts
        for ii in primitive.presentation_context_definition_results_list:
            presentation_context = PresentationContextItemAC()
            presentation_context.from_primitive(ii)
            self.variable_items.append(presentation_context)

        # Make user information
        user_information = UserInformationItem()
        user_information.from_primitive(primitive.user_information)
        self.variable_items.append(user_information)

    def to_primitive(self) -> "A_ASSOCIATE":
        """Return an A-ASSOCIATE (accept) primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_ASSOCIATE
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_ASSOCIATE

        primitive = A_ASSOCIATE()

        # The two reserved parameters at byte offsets 11 and 27 shall be set
        #    to called and calling AET byte the value shall not be
        #   tested when received (PS3.8 Table 9-17)
        primitive.called_ae_title = self.reserved_aet
        primitive.calling_ae_title = self.reserved_aec

        for item in self.variable_items:
            # Add application context
            if isinstance(item, ApplicationContextItem):
                primitive.application_context_name = item.application_context_name

            # Add presentation contexts
            elif isinstance(item, PresentationContextItemAC):
                primitive.presentation_context_definition_results_list.append(
                    item.to_primitive()
                )

            # Add user information
            elif isinstance(item, UserInformationItem):
                primitive.user_information = item.to_primitive()

        # 0x00 = Accepted
        primitive.result = 0x00

        return primitive

    @property
    def application_context_name(self) -> Optional[UID]:
        """Return the *Application Context Name*, if available.

        Returns
        -------
        pydicom.uid.UID or None
            The acceptor's *Application Context Name* or None if not available.
        """
        for item in self.variable_items:
            if isinstance(item, ApplicationContextItem):
                return item.application_context_name

        return None

    @property
    def called_ae_title(self) -> str:
        """Return the value sent in the *Called AE Title* reserved space.

        While the standard says this value should match the A-ASSOCIATE-RQ
        value there is no guarantee and this should not be used as a check
        value.

        Returns
        -------
        str
            The value the A-ASSOCIATE-AC sent in the *Called AE Title* reserved
            space. May be an empty string if unable to decode the value.
        """
        return self.reserved_aet

    @property
    def calling_ae_title(self) -> str:
        """Return the value sent in the *Calling AE Title* reserved space.

        While the standard says this value should match the A-ASSOCIATE-RQ
        value there is no guarantee and this should not be used as a check
        value.

        Returns
        -------
        bytes
            The value the A-ASSOCIATE-AC sent in the *Calling AE Title*
            reserved space. May be an empty string if unable to decode the
            value.
        """
        return self.reserved_aec

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return [
            ((6, 2), "protocol_version", self._wrap_unpack, [UNPACK_UINT2]),
            ((10, 16), "reserved_aet", self._wrap_bytes, []),  # Called AET
            ((26, 16), "reserved_aec", self._wrap_bytes, []),  # Calling AET
            ((74, None), "variable_items", self._wrap_generate_items, []),
        ]

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            ("protocol_version", PACK_UINT2, []),
            (None, self._wrap_pack, [0x0000, PACK_UINT2]),
            ("reserved_aet", self._wrap_encode_str, [16]),  # Called AET
            ("reserved_aec", self._wrap_encode_str, [16]),  # Calling AET
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
            ("variable_items", self._wrap_encode_items, []),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        length = 68
        for item in self.variable_items:
            length += len(item)

        return length

    @property
    def presentation_context(self) -> List[PresentationContextItemAC]:
        """Return a list of the Presentation Context Items.

        Returns
        -------
        list of pdu_items.PresentationContextItemAC
            The Presentation Context Items.
        """
        return [
            item
            for item in self.variable_items
            if isinstance(item, PresentationContextItemAC)
        ]

    @property
    def reserved_aec(self) -> str:
        """Get or set the A-ASSOCIATE-AC's reserved *Calling AE Title* value.

        Parameters
        ----------
        value : str or bytes
            The value at offset 27-42 in the A-ASSOCIATE-AC PDU, usually the
            same value as the *Calling AE Title*.

        Returns
        -------
        str
            The decoded value at offset 27-42 in the A-ASSOCIATE-AC PDU,
            usually the same value as the *Calling AE Title*, or ``''`` if
            unable to decode the value.
        """
        return self._reserved_aec

    @reserved_aec.setter
    def reserved_aec(self, value: Union[str, bytes]) -> None:
        """Set the A-ASSOCIATE-AC's *Calling AE Title* value."""
        if isinstance(value, bytes):
            # The value should not be tested - included that it's decodable
            try:
                value = decode_bytes(value).strip()
            except ValueError:
                value = ""

        self._reserved_aec = value

    @property
    def reserved_aet(self) -> str:
        """Get or set the A-ASSOCIATE-AC's reserved *Called AE Title* value.

        Parameters
        ----------
        value : str or bytes
            The value at offset 11-26 in the A-ASSOCIATE-AC PDU, usually the
            same value as the *Called AE Title*.

        Returns
        -------
        str
            The decoded value at offset 11-26 in the A-ASSOCIATE-AC PDU,
            usually the same value as the *Called AE Title*, or ``''`` if
            unable to decode the value.
        """
        return self._reserved_aet

    @reserved_aet.setter
    def reserved_aet(self, value: Union[str, bytes]) -> None:
        """Set the A-ASSOCIATE-AC's *Called AE Title* value."""
        if isinstance(value, bytes):
            # The value should not be tested - included that it's decodable
            try:
                value = decode_bytes(value).strip()
            except ValueError:
                value = ""

        self._reserved_aet = value

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = ["A-ASSOCIATE-AC PDU"]
        s.append("==================")
        s.append(f"  PDU type: 0x{self.pdu_type:02X}")
        s.append(f"  PDU length: {self.pdu_length} bytes")
        s.append(f"  Protocol version: {self.protocol_version}")
        s.append(f"  Reserved (Called AET):  {self._reserved_aet}")
        s.append(f"  Reserved (Calling AET): {self._reserved_aec}")
        s.append("")

        s.append("  Variable Items:")
        s.append("  ---------------")
        s.append("  * Application Context Item")
        s.append(f"    -  Context name: ={self.application_context_name}")
        s.append("  * Presentation Context Item(s):")

        for cx in self.presentation_context:
            item_str_list = str(cx).split("\n")
            s.append(f"    -  {item_str_list[0]}")
            for jj in item_str_list[1:-1]:
                s.append(f"       {jj}")

        s.append("  * User Information Item(s):")
        for item in cast(UserInformationItem, self.user_information).user_data:
            item_str_list = str(item).split("\n")
            s.append(f"    -  {item_str_list[0]}")
            for jj in item_str_list[1:-1]:
                s.append(f"       {jj}")

        return "\n".join(s)

    @property
    def user_information(self) -> Optional[UserInformationItem]:
        """Return the User Information Item, if available.

        Returns
        -------
        pdu_items.UserInformationItem or None
            The acceptor's User Information object or None, if not available.
        """
        for item in self.variable_items:
            if isinstance(item, UserInformationItem):
                return item

        return None


class A_ASSOCIATE_RJ(PDU):
    """An A-ASSOCIATE-RJ PDU.

    An A-ASSOCIATE-RJ PDU is sent by an association acceptor to indicate that
    association negotiation has been unsuccessful.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x03``).
    reason_diagnostic : int
        The *Reason/Diagnostic* field value.
    result : int
        The *Result* field value.
    source : int
        The *Source* field value.

    Notes
    -----
    An A-ASSOCIATE-RJ PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x03``)
    * PDU length (1)
    * Result (1)
    * Source (1)
    * Reason/Diagnostic (1)

    **Encoding**

    When encoded, an A-ASSOCIATE-RJ PDU has the following structure, taken
    from Table 9-21 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+-------------------+
    | Offset | Length      | Description       |
    +========+=============+===================+
    | 0      | 1           | PDU type          |
    +--------+-------------+-------------------+
    | 1      | 1           | Reserved          |
    +--------+-------------+-------------------+
    | 2      | 4           | PDU length        |
    +--------+-------------+-------------------+
    | 6      | 1           | Reserved          |
    +--------+-------------+-------------------+
    | 7      | 1           | Result            |
    +--------+-------------+-------------------+
    | 8      | 1           | Source            |
    +--------+-------------+-------------------+
    | 9      | 1           | Reason/diagnostic |
    +--------+-------------+-------------------+

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.4 <part08/sect_9.3.4.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["A_ASSOCIATE"] = None) -> None:
        """Initialise a new A-ASSOCIATE-RJ PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_ASSOCIATE
            The primitive to use to initialise the PDU.
        """
        self.result: Optional[int] = None
        self.source: Optional[int] = None
        self.reason_diagnostic: Optional[int] = None

        if primitive is not None:
            self.from_primitive(primitive)

    def from_primitive(self, primitive: "A_ASSOCIATE") -> None:
        """Setup the current PDU using an A-ASSOCIATE (reject) primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_ASSOCIATE
            The primitive to use to set the current PDU field values.
        """
        self.result = primitive.result
        self.source = primitive.result_source
        self.reason_diagnostic = primitive.diagnostic

    def to_primitive(self) -> "A_ASSOCIATE":
        """Return an A-ASSOCIATE (reject) primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_ASSOCIATE
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_ASSOCIATE

        primitive = A_ASSOCIATE()
        primitive.result = self.result
        primitive.result_source = self.source
        primitive.diagnostic = self.reason_diagnostic

        return primitive

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return [
            ((7, 1), "result", self._wrap_unpack, [UNPACK_UCHAR]),
            ((8, 1), "source", self._wrap_unpack, [UNPACK_UCHAR]),
            ((9, 1), "reason_diagnostic", self._wrap_unpack, [UNPACK_UCHAR]),
        ]

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("result", PACK_UCHAR, []),
            ("source", PACK_UCHAR, []),
            ("reason_diagnostic", PACK_UCHAR, []),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        return 4

    @property
    def reason_str(self) -> str:
        """Return a str describing the *Reason/Diagnostic* field value."""
        _reasons = {
            1: {
                1: "No reason given",
                2: "Application context name not supported",
                3: "Calling AE title not recognised",
                4: "Reserved",
                5: "Reserved",
                6: "Reserved",
                7: "Called AE title not recognised",
                8: "Reserved",
                9: "Reserved",
                10: "Reserved",
            },
            2: {1: "No reason given", 2: "Protocol version not supported"},
            3: {
                0: "Reserved",
                1: "Temporary congestion",
                2: "Local limit exceeded",
                3: "Reserved",
                4: "Reserved",
                5: "Reserved",
                6: "Reserved",
                7: "Reserved",
            },
        }

        if self.source not in _reasons:
            msg = "Invalid value in Source field in A-ASSOCIATE-RJ PDU"
            LOGGER.error(msg)
            raise ValueError(msg)

        if self.reason_diagnostic not in _reasons[self.source]:
            msg = "Invalid value in Reason field in A-ASSOCIATE-RJ PDU"
            LOGGER.error(msg)
            raise ValueError(msg)

        return _reasons[self.source][self.reason_diagnostic]

    @property
    def result_str(self) -> str:
        """Return a str describing the *Result* field value."""
        _results = {1: "Rejected (Permanent)", 2: "Rejected (Transient)"}

        if self.result not in _results:
            msg = "Invalid value in Result field in A-ASSOCIATE-RJ PDU"
            LOGGER.error(msg)
            raise ValueError(msg)

        return _results[self.result]

    @property
    def source_str(self) -> str:
        """Return a str describing the *Source* field value."""
        _sources = {
            1: "DUL service-user",
            2: "DUL service-provider (ACSE related)",
            3: "DUL service-provider (presentation related)",
        }

        if self.source not in _sources:
            msg = "Invalid value in Source field in A-ASSOCIATE-RJ PDU"
            LOGGER.error(msg)
            raise ValueError(msg)

        return _sources[self.source]

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = "A-ASSOCIATE-RJ PDU\n"
        s += "==================\n"
        s += f"  PDU type: 0x{self.pdu_type:02X}\n"
        s += f"  PDU length: {self.pdu_length} bytes\n"
        s += f"  Result: {self.result_str}\n"
        s += f"  Source: {self.source_str}\n"
        s += f"  Reason/Diagnostic: {self.reason_str}\n"

        return s


# Overridden _generate_items and _wrap_generate_items
class P_DATA_TF(PDU):
    """A P-DATA-TF PDU.

    A P-DATA-TF PDU is used once an association has been established to send
    DIMSE message data.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x04``).
    presentation_data_value_items : list of pdu.PresentationDataValueItem
        The *Presentation Data Value Item(s)* field value.

    Notes
    -----
    A P-DATA-TF PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x04``)
    * PDU length (1)
    * Presentation data value Item(s) (1 or more)

    **Encoding**

    When encoded, a P-DATA-TF PDU has the following structure, taken
    from Table 9-22 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+-------------------------------+
    | Offset | Length      | Description                   |
    +========+=============+===============================+
    | 0      | 1           | PDU type                      |
    +--------+-------------+-------------------------------+
    | 1      | 1           | Reserved                      |
    +--------+-------------+-------------------------------+
    | 2      | 4           | PDU length                    |
    +--------+-------------+-------------------------------+
    | 6      | Variable    | Presentation data value items |
    +--------+-------------+-------------------------------+

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.5 <part08/sect_9.3.5.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["P_DATA"] = None) -> None:
        """Initialise a new P-DATA-TF PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.P_DATA
            The primitive to use to initialise the PDU.
        """
        self.presentation_data_value_items: _PDVItem = []

        if primitive is not None:
            self.from_primitive(primitive)

    def from_primitive(self, primitive: "P_DATA") -> None:
        """Setup the current PDU using a P-DATA primitive.

        Parameters
        ----------
        primitive : pdu_primitives.P_DATA
            The primitive to use to set the current PDU field values.
        """
        for item in primitive.presentation_data_value_list:
            presentation_data_value = PresentationDataValueItem()
            presentation_data_value.presentation_context_id = item[0]
            presentation_data_value.presentation_data_value = item[1]
            self.presentation_data_value_items.append(presentation_data_value)

    def to_primitive(self) -> "P_DATA":
        """Return a P-DATA primitive from the current PDU.

        Returns
        -------
        pdu_primitives.P_DATA
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import P_DATA

        primitive = P_DATA()

        primitive.presentation_data_value_list = []
        for item in self.presentation_data_value_items:
            primitive.presentation_data_value_list.append(
                (
                    cast(int, item.presentation_context_id),
                    cast(bytes, item.presentation_data_value),
                )
            )
        return primitive

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return [
            ((6, None), "presentation_data_value_items", self._wrap_generate_items, [])
        ]

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            ("presentation_data_value_items", self._wrap_encode_items, []),
        ]

    @staticmethod
    def _generate_items(bytestream: bytes) -> Iterator[Tuple[int, bytes]]:
        """Yield the variable PDV item data from `bytestream`.

        Parameters
        ----------
        bytestream : bytes
            The encoded PDU variable item data.

        Yields
        ------
        int, bytes
            The PDV's Presentation Context ID as int, and the PDV item's
            encoded data as bytes.

        Notes
        -----
        **Encoding**
        When encoded, a Presentation Data Value Item has the following
        structure, taken from Table 9-23 (offset shown with Python
        indexing). The item is encoded using Big Endian, but the encoding of
        of the presentation data message fragments is dependent on the
        negotiated transfer syntax.

        +--------+-------------+-------------------------+
        | Offset | Length      | Description             |
        +========+=============+=========================+
        | 0      | 4           | Item length             |
        +--------+-------------+-------------------------+
        | 4      | 1           | Context ID              |
        +--------+-------------+-------------------------+
        | 5      | NN          | Presentation data value |
        +--------+-------------+-------------------------+

        References
        ----------
        * DICOM Standard, Part 8, :dcm:`Section
          9.3.5.1 <part08/sect_9.3.5.html#sect_9.3.5.1>`
        * DICOM Standard, Part 8,
          :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
        """
        offset = 0
        while bytestream[offset : offset + 1]:
            item_length = UNPACK_UINT4(bytestream[offset : offset + 4])[0]
            context_id = UNPACK_UCHAR(bytestream[offset + 4 : offset + 5])[0]
            data = bytestream[offset + 5 : offset + 4 + item_length]
            assert len(data) == item_length - 1
            yield context_id, data
            # Change `offset` to the start of the next PDV item
            offset += 4 + item_length

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        length = 0
        for item in self.presentation_data_value_items:
            length += len(item)

        return length

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = "P-DATA-TF PDU\n"
        s += "=============\n"
        s += f"  PDU type: 0x{self.pdu_type:02X}\n"
        s += f"  PDU length: {self.pdu_length} bytes\n"
        s += "\n"
        s += "  Presentation Data Value Item(s):\n"
        s += "  --------------------------------\n"

        for ii in self.presentation_data_value_items:
            item_str = f"{ii}"
            item_str_list = item_str.split("\n")
            s += f"  *  {item_str_list[0]}\n"
            for jj in item_str_list[1:]:
                s += f"     {jj}\n"

        return s

    def _wrap_generate_items(self, bytestream: bytes) -> _PDVItem:  # type: ignore
        """Return a list of PDV Items generated from `bytestream`.

        Parameters
        ----------
        bytestream : bytes
            The encoded presentation data value items.

        Returns
        -------
        list of pdu_items.PresentationDataValueItem
            The presentation data value items contained in `bytestream`.
        """
        item_list = []
        for context_id, data in self._generate_items(bytestream):
            pdv_item = PresentationDataValueItem()
            pdv_item.presentation_context_id = context_id
            pdv_item.presentation_data_value = data
            item_list.append(pdv_item)

        return item_list


class A_RELEASE_RQ(PDU):
    """An A-RELEASE-RQ PDU.

    An A-RELEASE-RQ PDU is used once an association has been established to
    initiate the release of the association.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x05``).

    Notes
    -----
    An A-RELEASE-RQ PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x05``)
    * PDU length (1, fixed value, 4)

    **Encoding**

    When encoded, an A-RELEASE-RQ PDU has the following structure, taken
    from Table 9-24 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+---------------+
    | Offset | Length      | Description   |
    +========+=============+===============+
    | 0      | 1           | PDU type      |
    +--------+-------------+---------------+
    | 1      | 1           | Reserved      |
    +--------+-------------+---------------+
    | 2      | 4           | PDU length    |
    +--------+-------------+---------------+
    | 6      | 4           | Reserved      |
    +--------+-------------+---------------+

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.6 <part08/sect_9.3.6.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["A_RELEASE"] = None) -> None:
        """Initialise a new A-RELEASE-RQ PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_RELEASE
            The primitive to use to initialise the PDU.
        """
        if primitive is not None:
            self.from_primitive(primitive)

    @staticmethod
    def from_primitive(primitive: "A_RELEASE") -> None:
        """Setup the current PDU using an A-RELEASE (request) primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_RELEASE
            The primitive to use to set the current PDU field values.
        """
        pass

    @staticmethod
    def to_primitive() -> "A_RELEASE":
        """Return an A-RELEASE (request) primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_RELEASE
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_RELEASE

        return A_RELEASE()

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return []

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        return 4

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = "A-RELEASE-RQ PDU\n"
        s += "================\n"
        s += f"  PDU type: 0x{self.pdu_type:02X}\n"
        s += f"  PDU length: {self.pdu_length} bytes\n"

        return s


class A_RELEASE_RP(PDU):
    """An A-RELEASE-RP PDU.

    An A-RELEASE-RP PDU is used once an association has been established to
    confirm the release of the association.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x06``).

    Notes
    -----
    An A-RELEASE-RP PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x06``)
    * PDU length (1, fixed value, ``0x00000004``)

    **Encoding**

    When encoded, an A-RELEASE-RP PDU has the following structure, taken
    from Table 9-25 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+---------------+
    | Offset | Length      | Description   |
    +========+=============+===============+
    | 0      | 1           | PDU type      |
    +--------+-------------+---------------+
    | 1      | 1           | Reserved      |
    +--------+-------------+---------------+
    | 2      | 4           | PDU length    |
    +--------+-------------+---------------+
    | 6      | 4           | Reserved      |
    +--------+-------------+---------------+

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.7 <part08/sect_9.3.7.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(self, primitive: Optional["A_RELEASE"] = None) -> None:
        """Initialise a new A-RELEASE-RP PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_RELEASE
            The primitive to use to initialise the PDU.
        """
        if primitive is not None:
            self.from_primitive(primitive)

    @staticmethod
    def from_primitive(primitive: "A_RELEASE") -> None:
        """Setup the current PDU using an A-release (response) primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_RELEASE
            The primitive to use to set the current PDU field values.
        """
        pass

    @staticmethod
    def to_primitive() -> "A_RELEASE":
        """Return an A-RELEASE (response) primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_RELEASE
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_RELEASE

        primitive = A_RELEASE()
        primitive.result = "affirmative"

        return primitive

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return []

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            (None, self._wrap_pack, [0x00, PACK_UINT4]),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        return 4

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = "A-RELEASE-RP PDU\n"
        s += "================\n"
        s += f"  PDU type: 0x{self.pdu_type:02X}\n"
        s += f"  PDU length: {self.pdu_length} bytes\n"

        return s


class A_ABORT_RQ(PDU):
    """An A-ABORT-RQ PDU.

    An A-ABORT-RQ PDU is used to abort the association.

    Attributes
    ----------
    pdu_type : int
        The *PDU Type* field value (``0x07``).
    reason_diagnostic : int
        The *Reason/Diagnostic* field value.
    source : int
        The *Source* field value.

    Notes
    -----
    An A-ABORT-RQ PDU requires the following parameters:

    * PDU type (1, fixed value, ``0x06``)
    * PDU length (1, fixed value, 4)
    * Source (1)
    * Reason/Diagnostic (1)

    **Encoding**

    When encoded, an A-ABORT-RQ PDU has the following structure, taken
    from Table 9-26 (offsets shown with Python indexing). PDUs are always
    encoded using Big Endian.

    +--------+-------------+-------------------+
    | Offset | Length      | Description       |
    +========+=============+===================+
    | 0      | 1           | PDU type          |
    +--------+-------------+-------------------+
    | 1      | 1           | Reserved          |
    +--------+-------------+-------------------+
    | 2      | 4           | PDU length        |
    +--------+-------------+-------------------+
    | 6      | 1           | Reserved          |
    +--------+-------------+-------------------+
    | 7      | 1           | Reserved          |
    +--------+-------------+-------------------+
    | 8      | 1           | Source            |
    +--------+-------------+-------------------+
    | 9      | 1           | Reason/Diagnostic |
    +--------+-------------+-------------------+

    References
    ----------
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.8 <part08/sect_9.3.8.html>`
    * DICOM Standard, Part 8,
      :dcm:`Section 9.3.1<part08/sect_9.3.html#sect_9.3.1>`
    """

    def __init__(
        self, primitive: Optional[Union["A_ABORT", "A_P_ABORT"]] = None
    ) -> None:
        """Initialise a new A-ABORT-RQ PDU.

        Parameters
        ----------
        primitive : pynetdicom.pdu_primitive.A_ABORT
            The primitive to use to initialise the PDU.
        """
        self.source: Optional[int] = None
        self.reason_diagnostic: Optional[int] = None

        if primitive is not None:
            self.from_primitive(primitive)

    def from_primitive(self, primitive: _AbortType) -> None:
        """Setup the current PDU using an A-ABORT or A-P-ABORT primitive.

        Parameters
        ----------
        primitive : pdu_primitives.A_ABORT or pdu_primitives.A_P_ABORT
            The primitive to use to set the current PDU field values.
        """
        from pynetdicom.pdu_primitives import A_ABORT, A_P_ABORT

        # User initiated abort
        if isinstance(primitive, A_ABORT):
            # The reason field shall be 0x00 when the source is DUL
            # service-user
            self.reason_diagnostic = 0
            self.source = primitive.abort_source

        # User provider primitive abort
        elif isinstance(primitive, A_P_ABORT):
            self.reason_diagnostic = primitive.provider_reason
            self.source = 2

    def to_primitive(self) -> _AbortType:
        """Return an A-ABORT or A-P-ABORT primitive from the current PDU.

        Returns
        -------
        pdu_primitives.A_ABORT or pdu_primitives.A_P_ABORT
            The primitive representation of the current PDU.
        """
        from pynetdicom.pdu_primitives import A_ABORT, A_P_ABORT

        primitive: _AbortType
        if self.source == 0x02:
            # User provider primitive abort
            primitive = A_P_ABORT()
            primitive.provider_reason = self.reason_diagnostic
        else:
            # User initiated abort and undefined abort source
            primitive = A_ABORT()
            primitive.abort_source = self.source

        return primitive

    @property
    def _decoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of ((offset, length), attr_name, callable, [args]), where

            - offset is the byte offset to start at
            - length is how many bytes to slice (if None then will slice to the
              end of the data),
            - attr_name is the name of the attribute corresponding to the field
            - callable is a decoding function that returns the decoded value,
            - args is a list of arguments to pass callable.
        """
        return [
            ((8, 1), "source", self._wrap_unpack, [UNPACK_UCHAR]),
            ((9, 1), "reason_diagnostic", self._wrap_unpack, [UNPACK_UCHAR]),
        ]

    @property
    def _encoders(self) -> Any:
        """Return an iterable of tuples that contain field decoders.

        Returns
        -------
        list of tuple
            A list of (attr_name, callable, [args]), where

            - attr_name is the name of the attribute corresponding to the field
            - callable is an encoding function that returns bytes
            - args is a list of arguments to pass callable.
        """
        return [
            ("pdu_type", PACK_UCHAR, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("pdu_length", PACK_UINT4, []),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            (None, self._wrap_pack, [0x00, PACK_UCHAR]),
            ("source", PACK_UCHAR, []),
            ("reason_diagnostic", PACK_UCHAR, []),
        ]

    @property
    def pdu_length(self) -> int:
        """Return the *PDU Length* field value as an int."""
        return 4

    def __str__(self) -> str:
        """Return a string representation of the PDU."""
        s = "A-ABORT PDU\n"
        s += "===========\n"
        s += f"  PDU type: 0x{self.pdu_type:02X}\n"
        s += f"  PDU length: {self.pdu_length} bytes\n"
        s += f"  Abort Source: {self.source_str}\n"
        s += f"  Reason/Diagnostic: {self.reason_str}\n"

        return s

    @property
    def source_str(self) -> str:
        """Return a str description of the *Source* field value."""
        _sources = {0: "DUL service-user", 1: "Reserved", 2: "DUL service-provider"}
        if self.source not in _sources:
            LOGGER.warning(f"Invalid A-ABORT-RQ 'Source' {self.source}")
            return "(no value available)"

        return _sources[self.source]

    @property
    def reason_str(self) -> str:
        """Return a str description of the *Reason/Diagnostic* field value."""
        if self.source == 2:
            _reason_str = {
                0: "No reason given",
                1: "Unrecognised PDU",
                2: "Unexpected PDU",
                3: "Reserved",
                4: "Unrecognised PDU parameter",
                5: "Unexpected PDU parameter",
                6: "Invalid PDU parameter value",
            }
            if self.reason_diagnostic not in _reason_str:
                LOGGER.warning(
                    "Invalid A-ABORT-RQ 'Reason/Diagnostic' "
                    f"{self.reason_diagnostic}"
                )
                return "(no value available)"

            return _reason_str[self.reason_diagnostic]

        return "(no value available)"


# PDUs indexed by their class
PDU_TYPES = {
    A_ASSOCIATE_RQ: 0x01,
    A_ASSOCIATE_AC: 0x02,
    A_ASSOCIATE_RJ: 0x03,
    P_DATA_TF: 0x04,
    A_RELEASE_RQ: 0x05,
    A_RELEASE_RP: 0x06,
    A_ABORT_RQ: 0x07,
}



"""Implementation of the Transport Service."""

from copy import deepcopy
from datetime import datetime
import gc
import logging
import queue
import select
import socket
from socketserver import TCPServer, ThreadingMixIn, BaseRequestHandler

try:
    import ssl

    _HAS_SSL = True
except ImportError:
    _HAS_SSL = False
import threading
from typing import (
    TYPE_CHECKING,
    Optional,
    Any,
    Tuple,
    cast,
    List,
    Dict,
    Callable,
    Union,
)

from pynetdicom import evt, _config
from pynetdicom._globals import MODE_ACCEPTOR, BIND_ADDRESS
from pynetdicom._handlers import (
    standard_dimse_recv_handler,
    standard_dimse_sent_handler,
    standard_pdu_recv_handler,
    standard_pdu_sent_handler,
)
from pynetdicom.pdu_primitives import A_ASSOCIATE
from pynetdicom.presentation import PresentationContext

if TYPE_CHECKING:  # pragma: no cover
    from pynetdicom.ae import ApplicationEntity
    from pynetdicom.association import Association
    from pynetdicom.dul import _QueueType


LOGGER = logging.getLogger("pynetdicom.transport")


class T_CONNECT:
    """A TRANSPORT CONNECTION primitive

    .. versionadded:: 2.0

    Attributes
    ----------
    request : pynetdicom.pdu_primitives.A_ASSOCIATE
        The A-ASSOCIATE (request) primitive that generated the TRANSPORT CONNECTION
        primitive.
    """

    def __init__(self, request: "A_ASSOCIATE") -> None:
        """Create a new TRANSPORT CONNECTION primitive.

        Parameters
        ----------
        request : pynetdicom.pdu_primitives.A_ASSOCIATE
            The A-ASSOCIATE (request) primitive to use when making a connection with
            a peer.
        """
        self._result = ""
        self.request = request

        if not isinstance(request, A_ASSOCIATE):
            raise TypeError(
                f"'request' must be 'pynetdicom.pdu_primitives.A_ASSOCIATE', not "
                f"'{request.__class__.__name__}'"
            )

    @property
    def address(self) -> Tuple[str, int]:
        """Return the peer's ``(str: IP address, int: port)``."""
        return cast(Tuple[str, int], self.request.called_presentation_address)

    @property
    def result(self) -> str:
        """Return the result of the connection attempt as :class:`str`.

        Parameters
        ----------
        str
            The result of the connection attempt, ``"Evt2"`` if the connection
            succeeded, ``"Evt17"`` if it failed.

        Returns
        -------
        str
            The result of the connection attempt, ``"Evt2"`` if the connection
            succeeded, ``"Evt17"`` if it failed.
        """
        if self._result == "":
            raise ValueError("A connection attempt has not yet been made")

        return self._result

    @result.setter
    def result(self, value: str) -> None:
        if value not in ("Evt2", "Evt17"):
            raise ValueError(f"Invalid connection result '{value}'")

        self._result = value


class AssociationSocket:
    """A wrapper for a `socket
    <https://docs.python.org/3/library/socket.html#socket-objects>`_ object.

    .. versionadded:: 1.2

    Provides an interface for ``socket`` that is integrated
    nicely with an :class:`~pynetdicom.association.Association` instance
    and the state machine.

    Attributes
    ----------
    select_timeout : float or None
        The timeout (in seconds) that :func:`select.select` calls in
        :meth:`ready` will block for (default ``0.5``). A value of ``0``
        specifies a poll and never blocks. A value of ``None`` blocks until a
        connection is ready.
    socket : socket.socket or None
        The wrapped socket, will be ``None`` if :meth:`close` is called.
    """

    def __init__(
        self,
        assoc: "Association",
        client_socket: Optional[socket.socket] = None,
        address: Tuple[str, int] = BIND_ADDRESS,
    ) -> None:
        """Create a new :class:`AssociationSocket`.

        .. versionchanged:: 2.0

            The default for *address* was changed to ``("127.0.0.1", 0)``

        Parameters
        ----------
        assoc : association.Association
            The :class:`~pynetdicom.association.Association` instance that will
            be using the socket to communicate.
        client_socket : socket.socket, optional
            The ``socket.socket`` to wrap,
            if not supplied then a new socket will be created instead.
        address : 2-tuple, optional
            If *client_socket* is ``None`` then this is the ``(host, port)`` to
            bind the newly created socket to, which by default will be
            ``('127.0.0.1', 0)``.
        """
        self._assoc = assoc

        if client_socket is not None and address != BIND_ADDRESS:
            LOGGER.warning(
                "AssociationSocket instantiated with both a 'client_socket' "
                "and bind 'address'. The original socket will not be rebound"
            )

        self._ready = threading.Event()
        self.socket: Optional[socket.socket]

        if client_socket is None:
            self.socket = self._create_socket(address)
            self._is_connected = False
        else:
            self.socket = client_socket
            self._is_connected = True
            self._ready.set()
            # Evt5: Transport connection indication
            self.event_queue.put("Evt5")

        self._tls_args: Optional[Tuple["ssl.SSLContext", str]] = None
        self.select_timeout = 0.5

    @property
    def assoc(self) -> "Association":
        """Return the parent :class:`~pynetdicom.association.Association`
        instance.
        """
        return self._assoc

    def close(self) -> None:
        """Close the connection to the peer and shutdown the socket.

        Sets :attr:`AssociationSocket.socket` to ``None`` once complete.

        **Events Emitted**

        - Evt17: Transport connection closed
        """
        if self.socket is None or self._is_connected is False:
            return

        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass

        self.socket.close()
        self.socket = None
        self._is_connected = False
        # Evt17: Transport connection closed
        self.event_queue.put("Evt17")

    def connect(self, primitive: T_CONNECT) -> None:
        """Try and connect to a remote at `address`.

        .. versionchanged:: 2.0

            Changed to take a :class:`~pynetdicom.transport.T_CONNECT` primitive rather
            than an address tuple.

        Parameters
        ----------
        primitive : pynetdicom.transport.T_CONNECT
            The TRANSPORT CONNECT primitive to use when connecting to a peer.
        """
        if self.socket is None:
            self.socket = self._create_socket()

        try:
            if self.tls_args:
                context, server_hostname = self.tls_args
                self.socket = cast(
                    socket.socket,
                    context.wrap_socket(
                        self.socket,
                        server_side=False,
                        server_hostname=server_hostname,
                    ),
                )
            # Set ae connection timeout
            self.socket.settimeout(self.assoc.connection_timeout)
            # Try and connect to remote at (address, port)
            #   raises socket.error if connection refused
            self.socket.connect(primitive.address)
            # Clear ae connection timeout
            self.socket.settimeout(None)
            # Trigger event - connection open
            evt.trigger(self.assoc, evt.EVT_CONN_OPEN, {"address": primitive.address})
            self._is_connected = True
            # Evt2: Transport connection confirmation
            primitive.result = "Evt2"
            self.provider_queue.put(primitive)
        except OSError as exc:
            # Log connection failure
            LOGGER.error("Association request failed: unable to connect to remote")
            LOGGER.error(f"TCP Initialisation Error: {exc}")
            # Log exception if TLS issue to help with troubleshooting
            if isinstance(exc, ssl.SSLError):
                LOGGER.exception(exc)

            # Don't be tempted to replace this with a self.close() call -
            #   it doesn't work because `_is_connected` is False
            if self.socket:
                try:
                    self.socket.shutdown(socket.SHUT_RDWR)
                except Exception:
                    pass
                self.socket.close()
                self.socket = None

            primitive.result = "Evt17"
            self.provider_queue.put(primitive)
        finally:
            self._ready.set()

    def _create_socket(self, address: Tuple[str, int] = BIND_ADDRESS) -> socket.socket:
        """Create a new IPv4 TCP socket and set it up for use.

        *Socket Options*

        - :attr:`socket.SO_REUSEADDR` is 1
        - :meth:`socket.settimeout` sets the Association's
          :attr:`~pynetdicom.association.Associaton.network_timeout` value.

        Parameters
        ----------
        address : 2-tuple, optional
            The ``(host: str, port: int)`` to bind the socket to. By default the socket
            is bound to ``("127.0.0.1", 0)``, i.e. the first available port.

        Returns
        -------
        socket.socket
            A bound and unconnected socket instance.
        """
        # AF_INET: IPv4, SOCK_STREAM: TCP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # SO_REUSEADDR: reuse the socket in TIME_WAIT state without
        #   waiting for its natural timeout to expire
        #   Allows local address reuse
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # If no timeout is set then recv() will block forever if
        #   the connection is kept alive with no data sent
        if self.assoc.network_timeout is not None:
            sock.settimeout(self.assoc.network_timeout)

        sock.bind(address)

        self._is_connected = False

        return sock

    @property
    def event_queue(self) -> "queue.Queue[str]":
        """Return the :class:`~pynetdicom.association.Association`'s service event
        queue.
        """
        return self.assoc.dul.event_queue

    def get_local_addr(self, host: Tuple[str, int] = ("10.255.255.255", 1)) -> str:
        """Return an address for the local computer as :class:`str`.

        Parameters
        ----------
        host : Tuple[str, int]
            The host's ``(addr: str, port: int)`` when trying to determine the local
            address.
        """
        # Solution from https://stackoverflow.com/a/28950776
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            try:
                # We use `host` to allow unit testing
                sock.connect(host)
                addr: str = sock.getsockname()[0]
            except Exception:
                addr = "127.0.0.1"

        return addr

    @property
    def provider_queue(self) -> "_QueueType":
        """Return the :class:`~pynetdicom.association.Association`'s service provider
        queue.
        """
        return self.assoc.dul.to_provider_queue

    @property
    def ready(self) -> bool:
        """Return ``True`` if there is data available to be read.

        *Events Emitted*

        - None
        - Evt17: Transport connection closed

        Returns
        -------
        bool
            ``True`` if the socket has data ready to be read, ``False``
            otherwise.
        """
        if self.socket is None or self._is_connected is False:
            return False

        try:
            # Use a timeout of 0 so we get an "instant" result
            ready, _, _ = select.select([self.socket], [], [], 0)
        except (socket.error, socket.timeout, ValueError):
            # Evt17: Transport connection closed
            self.event_queue.put("Evt17")
            return False

        # An SSLSocket may have buffered data available that `select`
        # is unaware of - see #528
        if _HAS_SSL and isinstance(self.socket, ssl.SSLSocket):
            return bool(ready) or bool(self.socket.pending())

        return bool(ready)

    def recv(self, nr_bytes: int) -> bytearray:
        """Read `nr_bytes` from the socket.

        *Events Emitted*

        - None

        Parameters
        ----------
        nr_bytes : int
            The number of bytes to attempt to read from the socket.

        Returns
        -------
        bytearray
            The data read from the socket.
        """
        self.socket = cast(socket.socket, self.socket)
        bytestream = bytearray()
        nr_read = 0
        # socket.recv() returns when the network buffer has been emptied
        #   not necessarily when the number of bytes requested have been
        #   read. Its up to us to keep calling recv() until we have all the
        #   data we want
        # **BLOCKING** until either all the data is read or an error occurs
        while nr_read < nr_bytes:
            # Python docs recommend reading a relatively small power of 2
            #   such as 4096
            bufsize = 4096
            if (nr_bytes - nr_read) < bufsize:
                bufsize = nr_bytes - nr_read

            bytes_read = self.socket.recv(bufsize)

            # If socket.recv() reads 0 bytes then the connection has been
            #   broken, so return what we have so far
            if not bytes_read:
                return bytestream

            bytestream.extend(bytes_read)
            nr_read += len(bytes_read)

        return bytestream

    def send(self, bytestream: bytes) -> None:
        """Try and send the data in `bytestream` to the remote.

        *Events Emitted*

        - None
        - Evt17: Transport connected closed.

        Parameters
        ----------
        bytestream : bytes
            The data to send to the remote.
        """
        self.socket = cast(socket.socket, self.socket)
        total_sent = 0
        length_data = len(bytestream)
        try:
            while total_sent < length_data:
                # Returns the number of bytes sent
                nr_sent = self.socket.send(bytestream[total_sent:])
                total_sent += nr_sent

            evt.trigger(self.assoc, evt.EVT_DATA_SENT, {"data": bytestream})
        except (socket.error, socket.timeout):
            # Evt17: Transport connection closed
            self.event_queue.put("Evt17")

    def __str__(self) -> str:
        """Return the string output for ``socket``."""
        return self.socket.__str__()

    @property
    def tls_args(self) -> Optional[Tuple["ssl.SSLContext", str]]:
        """Get or set the TLS context and hostname.

        Parameters
        ----------
        tls_args : Tuple[ssl.SSLContext, str] or None
            If the socket should be wrapped by TLS then this is
            ``(context, hostname)``, where *context* is a
            :class:`ssl.SSLContext` that will be used to wrap the socket and
            *hostname* is the value to use for the *server_hostname* keyword
            argument for :meth:`SSLContext.wrap_socket()
            <ssl.SSLContext.wrap_socket>`.

        Returns
        -------
        Optional[Tuple[ssl.SSLContext, str]]
        """
        return self._tls_args

    @tls_args.setter
    def tls_args(self, tls_args: Optional[Tuple["ssl.SSLContext", str]]) -> None:
        """Set the TLS arguments for the socket."""
        if not _HAS_SSL:
            raise RuntimeError("Your Python installation lacks support for SSL")

        self._tls_args = tls_args


class RequestHandler(BaseRequestHandler):
    """Connection request handler for the ``AssociationServer``.

    .. versionadded:: 1.2

    Attributes
    ----------
    client_address : 2-tuple
        The ``(host, port)`` of the remote.
    request : socket.socket
        The (unaccepted) client socket.
    server : transport.AssociationServer or transport.ThreadedAssociationServer
        The server that received the connection request.
    """

    server: "AssociationServer"

    @property
    def ae(self) -> "ApplicationEntity":
        """Return the server's parent AE."""
        return self.server.ae

    def handle(self) -> None:
        """Handle an association request.

        * Creates a new Association acceptor instance and configures it.
        * Sets the Association's socket to the request's socket.
        * Starts the Association reactor.
        """
        assoc = self._create_association()

        # Trigger must be after binding the events
        evt.trigger(assoc, evt.EVT_CONN_OPEN, {"address": self.client_address})

        assoc.start()

    @property
    def local(self) -> Tuple[str, int]:
        """Return a 2-tuple of the local server's ``(host, port)`` address."""
        return self.server.server_address

    @property
    def remote(self) -> Tuple[str, int]:
        """Return a 2-tuple of the remote client's ``(host, port)`` address."""
        return cast(Tuple[str, int], self.client_address)

    def _create_association(self) -> "Association":
        """Create an :class:`Association` object for the current request.

        .. versionadded:: 1.5
        """
        from pynetdicom.association import Association

        assoc = Association(self.ae, MODE_ACCEPTOR)
        assoc._server = self.server

        # Set the thread name
        timestamp = datetime.strftime(datetime.now(), "%Y%m%d%H%M%S")
        assoc.name = f"AcceptorThread@{timestamp}"

        sock = AssociationSocket(assoc, client_socket=self.request)
        assoc.set_socket(sock)

        # Association Acceptor object -> local AE
        assoc.acceptor.maximum_length = self.ae.maximum_pdu_size
        assoc.acceptor.ae_title = self.server.ae_title
        assoc.acceptor.address = self.local[0]
        assoc.acceptor.port = self.local[1]
        assoc.acceptor.implementation_class_uid = self.ae.implementation_class_uid
        assoc.acceptor.implementation_version_name = self.ae.implementation_version_name
        assoc.acceptor.supported_contexts = deepcopy(self.server.contexts)

        # Association Requestor object -> remote AE
        assoc.requestor.address = self.remote[0]
        assoc.requestor.port = self.remote[1]

        # Bind events to handlers
        for event in self.server._handlers:
            # Intervention events
            if event.is_intervention and self.server._handlers[event]:

                assoc.bind(event, *self.server._handlers[event])
            elif isinstance(event, evt.NotificationEvent):
                # List[Tuple[Callable, Optional[List[Any]]]]
                for handler in self.server._handlers[event]:
                    handler = cast(evt._HandlerBase, handler)
                    assoc.bind(event, handler[0], handler[1])

        return assoc


class AssociationServer(TCPServer):
    """An Association server implementation.

    .. versionadded:: 1.2

    .. versionchanged:: 1.5

        Added `request_handler` keyword parameter.

    Any attempts to connect will be assumed to be from association requestors.

    The server should be started with
    :meth:`serve_forever(poll_interval)<AssociationServer.serve_forever>`,
    where *poll_interval* is the timeout (in seconds) that the
    :func:`select.select` call will block for (default ``0.5``). A value of
    ``0`` specifies a poll and never blocks. A value of ``None`` blocks until
    a connection is ready.

    Attributes
    ----------
    ae : ae.ApplicationEntity
        The parent AE that is running the server.
    request_queue_size : int
        Default ``5``.
    server_address : Tuple[str, int]
        The ``(host: str, port: int)`` that the server is running on.
    """

    def __init__(
        self,
        ae: "ApplicationEntity",
        address: Tuple[str, int],
        ae_title: str,
        contexts: List[PresentationContext],
        ssl_context: Optional["ssl.SSLContext"] = None,
        evt_handlers: List[evt.EventHandlerType] = None,
        request_handler: Optional[Callable[..., BaseRequestHandler]] = None,
    ) -> None:
        """Create a new :class:`AssociationServer`, bind a socket and start
        listening.

        Parameters
        ----------
        ae : ae.ApplicationEntity
            The parent AE that's running the server.
        address : Tuple[str, int]
            The ``(host: str, port: int)`` that the server should run on.
        ae_title : str
            The AE title of the SCP.
        contexts : list of presentation.PresentationContext
            The SCPs supported presentation contexts.
        ssl_context : ssl.SSLContext, optional
            If TLS is to be used then this should be the
            :class:`ssl.SSLContext` used to wrap the client sockets, otherwise
            if ``None`` then no TLS will be used (default).
        evt_handlers : list of 2- or 3-tuple, optional
            A list of ``(event, callable)`` or ``(event, callable, args)``,
            the *callable* function to run when *event* occurs and the
            optional extra *args* to pass to the callable.
        request_handler : type
            The request handler class; an instance of this class
            is created for each request. Should be a subclass of
            :class:`~socketserver.BaseRequestHandler`.
        """
        self.ae = ae
        self.ae_title = ae_title
        self.contexts = contexts
        self.ssl_context = ssl_context
        self.allow_reuse_address = True
        self.server_address: Tuple[str, int] = address
        self.socket: Optional[socket.socket] = None  # type: ignore[assignment]

        request_handler = request_handler or RequestHandler

        super().__init__(address, request_handler, bind_and_activate=True)

        self.timeout = 60

        # Stores all currently bound event handlers so future
        #   Associations can be bound
        self._handlers: Dict[
            evt.EventType,
            Union[
                List[Tuple[Callable, Optional[List[Any]]]],
                Tuple[Callable, Optional[List[Any]]],
            ],
        ] = {}
        self._bind_defaults()

        # Bind the functions to their events
        for evt_hh_args in evt_handlers or ():
            self.bind(*evt_hh_args)

        self._gc = [0, 59]

    def bind(
        self, event: evt.EventType, handler: Callable, args: Optional[List[Any]] = None
    ) -> None:
        """Bind a callable `handler` to an `event`.

        .. versionadded:: 1.3

        .. versionchanged:: 1.5

            Added `args` keyword parameter.

        Parameters
        ----------
        event : namedtuple
            The event to bind the function to.
        handler : callable
            The function that will be called if the event occurs.
        args : list, optional
            Optional extra arguments to be passed to the handler (default:
            no extra arguments passed to the handler).
        """
        evt._add_handler(event, self._handlers, (handler, args))

        # Bind our child Association events
        for assoc in self.active_associations:
            assoc.bind(event, handler, args)

    def _bind_defaults(self) -> None:
        """Bind the default event handlers."""
        # Intervention event handlers
        for event in evt._INTERVENTION_EVENTS:
            handler = evt.get_default_handler(event)
            self.bind(event, handler)

        # Notification event handlers
        if _config.LOG_HANDLER_LEVEL == "standard":
            self.bind(evt.EVT_DIMSE_RECV, standard_dimse_recv_handler)
            self.bind(evt.EVT_DIMSE_SENT, standard_dimse_sent_handler)
            self.bind(evt.EVT_PDU_RECV, standard_pdu_recv_handler)
            self.bind(evt.EVT_PDU_SENT, standard_pdu_sent_handler)

    @property
    def active_associations(self) -> List["Association"]:
        """Return the server's running
        :class:`~pynetdicom.association.Association` acceptor instances
        """
        # Find all AcceptorThreads with `_server` as self
        threads = cast(
            List["Association"],
            [tt for tt in threading.enumerate() if "AcceptorThread" in tt.name],
        )
        return [tt for tt in threads if tt._server is self]

    def get_events(self) -> List[evt.EventType]:
        """Return a list of currently bound events.

        .. versionadded:: 1.3
        """
        return sorted(self._handlers.keys(), key=lambda x: x.name)

    def get_handlers(self, event: evt.EventType) -> evt.HandlerArgType:
        """Return handlers bound to a specific `event`.

        .. versionadded:: 1.3

        .. versionchanged:: 1.5

            Returns a 2-tuple of (callable, args) or list of 2-tuple.

        Parameters
        ----------
        event : namedtuple
            The event bound to the handlers.

        Returns
        -------
        2-tuple of (callable, args), list of 2-tuple
            If the event is a notification event then returns a list of
            2-tuples containing the callable functions bound to `event` and
            the arguments passed to the callable as ``(callable, args)``. If
            the event is an intervention event then returns either a 2-tuple of
            (callable, args) if a handler is bound to the event or
            ``(None, None)`` if no handler has been bound.
        """
        if event not in self._handlers:
            return []

        return self._handlers[event]

    def get_request(self) -> Tuple[socket.socket, Tuple[str, int]]:
        """Handle a connection request.

        If :attr:`~AssociationServer.ssl_context` is set then the client socket
        will be wrapped using
        :meth:`SSLContext.wrap_socket()<ssl.SSLContext.wrap_socket>`.

        Returns
        -------
        client_socket : socket.socket
            The connection request.
        address : 2-tuple
            The client's address as ``(host, port)``.
        """
        self.socket = cast(socket.socket, self.socket)
        client_socket, address = self.socket.accept()
        if self.ssl_context:
            client_socket = self.ssl_context.wrap_socket(
                client_socket, server_side=True
            )

        return client_socket, address

    def process_request(
        self,
        request: Union[socket.socket, Tuple[bytes, socket.socket]],
        client_address: Union[Tuple[str, int], str],
    ) -> None:
        """Process a connection request"""
        # Calls request_handler(request, client_address, self)
        self.finish_request(request, client_address)

    def server_bind(self) -> None:
        """Bind the socket and set the socket options.

        - ``socket.SO_REUSEADDR`` is set to ``1``
        - socket.settimeout is used to set to
          :attr:`AE.network_timeout
          <pynetdicom.ae.ApplicationEntity.network_timeout>` unless the
          value is ``None`` in which case it will be left unset.
        """
        self.socket = cast(socket.socket, self.socket)
        # SO_REUSEADDR: reuse the socket in TIME_WAIT state without
        #   waiting for its natural timeout to expire
        #   Allows local address reuse
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        # If no timeout is set then recv() will block forever if
        #   the connection is kept alive with no data sent
        if self.ae.network_timeout is not None:
            self.socket.settimeout(self.ae.network_timeout)

        # Bind the socket to an (address, port)
        #   If address is '' then the socket is reachable by any
        #   address the machine may have, otherwise is visible only on that
        #   address
        self.socket.bind(self.server_address)
        self.server_address = self.socket.getsockname()

    def server_close(self) -> None:
        """Close the server."""
        self.socket = cast(socket.socket, self.socket)
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except socket.error:
            pass

        self.socket.close()

    def service_actions(self) -> None:
        """Called by the serve_forever() loop"""
        # For whatever reason dead Association threads aren't being garbage
        #   collected so do it manually when a request is received
        if self._gc[0] == self._gc[1]:
            gc.collect()
            self._gc[0] = 0
            return

        self._gc[0] += 1

    def shutdown(self) -> None:
        """Completely shutdown the server and close it's socket."""
        super().shutdown()
        self.server_close()
        self.ae._servers.remove(cast("ThreadedAssociationServer", self))

    @property
    def ssl_context(self) -> Optional["ssl.SSLContext"]:
        """Return the :class:`ssl.SSLContext` (if available).

        Parameters
        ----------
        context : ssl.SSLContext or None
            If TLS is to be used then this should be the
            :class:`ssl.SSLContext` used to wrap the client sockets, otherwise
            if ``None`` then no TLS will be used (default).
        """
        return self._ssl_context

    @ssl_context.setter
    def ssl_context(self, context: Optional["ssl.SSLContext"]) -> None:
        """Set the SSL context for the socket."""
        if not _HAS_SSL:
            raise RuntimeError("Your Python installation lacks support for SSL")

        self._ssl_context = context

    def unbind(self, event: evt.EventType, handler: Callable) -> None:
        """Unbind a callable `handler` from an `event`.

        .. versionadded:: 1.3

        Parameters
        ----------
        event : 3-tuple
            The event to unbind the function from.
        handler : callable
            The function that will no longer be called if the event occurs.
        """
        evt._remove_handler(event, self._handlers, handler)

        # Unbind from our child Association events
        for assoc in self.active_associations:
            assoc.unbind(event, handler)


class ThreadedAssociationServer(ThreadingMixIn, AssociationServer):
    """An :class:`AssociationServer` suitable for threading.

    .. versionadded:: 1.2
    """

    def process_request_thread(
        self,
        request: Union[socket.socket, Tuple[bytes, socket.socket]],
        client_address: Union[Tuple[str, int], str],
    ) -> None:
        """Process a connection request."""
        # pylint: disable=broad-except
        try:
            self.finish_request(request, client_address)
        except Exception:
            self.handle_error(request, client_address)
            self.shutdown_request(request)



"""Validation functions used by pynetdicom"""

import logging
from typing import Tuple
import unicodedata

from pydicom.uid import UID


LOGGER = logging.getLogger("pynetdicom._validators")


def validate_ae(value: str) -> Tuple[bool, str]:
    """Return ``True`` if `value` is a conformant **AE** value.

    An **AE** value:

    * Must be no more than 16 characters
    * Leading and trailing spaces are not significant
    * May only use ASCII characters, excluding ``0x5C`` (backslash) and all
      control characters

    Parameters
    ----------
    value : str
        The **AE** value to check.

    Returns
    -------
    Tuple[bool, str]
        A tuple of (bool, str), with the first item being ``True`` if the
        value is conformant to the DICOM Standard and ``False`` otherwise and
        the second item being a short description of why the validation failed
        or ``''`` if validation was successful.
    """
    if not isinstance(value, str):
        return False, "must be str"

    if len(value) > 16:
        return False, "must not exceed 16 characters"

    # All characters use ASCII
    if not value.isascii():
        return False, "must only contain ASCII characters"

    # Unicode category: 'Cc' is control characters
    invalid = [c for c in value if unicodedata.category(c)[0] == "C"]
    if invalid or "\\" in value:
        return False, "must not contain control characters or backslashes"

    return True, ""


def validate_ui(value: UID) -> Tuple[bool, str]:

    from pynetdicom import _config

    if not isinstance(value, str):
        return False, "must be pydicom.uid.UID"

    value = UID(value)

    if _config.ENFORCE_UID_CONFORMANCE:
        if value.is_valid:
            return True, ""

        return False, "UID is non-conformant"

    if not 0 < len(value):
        return False, "must not be an empty str"

    if not len(value) < 65:
        return False, "must not exceed 64 characters"

    return True, ""
